from __future__ import annotations

__all__ = (
    "Bound",
    "INF",
)

from dataclasses import dataclass


@dataclass(frozen=True)
class Bound:
    """An integer but sometimes also possibly infinite (None)"""

    v: int | None

    def __post_init__(self, /) -> None:
        if self.v is not None and self.v < 0:
            raise ValueError(f"Invalid bound: {self.v!r}")

    def __repr__(self, /) -> str:
        return f"Bound({self.v!r})"

    def __str__(self, /) -> str:
        if self.v is None:
            # This only happens for an unlimited upper bound
            return ""
        return str(self.v)

    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.v == other.v

    def __hash__(self, /) -> int:
        return hash(self.v)

    def __lt__(self, other: Bound, /) -> bool:
        if self.v is None:
            return False
        if other.v is None:
            return True
        return self.v < other.v

    def __ge__(self, other: Bound, /) -> bool:
        return not self < other

    def __mul__(self, other: Bound, /) -> Bound:
        """Multiply this bound by another"""
        if Bound(0) in (self, other):
            return Bound(0)
        if self.v is None or other.v is None:
            return INF
        return Bound(self.v * other.v)

    def __add__(self, other: Bound, /) -> Bound:
        """Add this bound to another"""
        if self.v is None or other.v is None:
            return INF
        return Bound(self.v + other.v)

    def __sub__(self, other: Bound, /) -> Bound:
        """
        Subtract another bound from this one.
        Caution: this operation is not meaningful for all bounds.
        """
        if other.v is None:
            if self.v is not None:
                raise ArithmeticError(f"Can't subtract {other!r} from {self!r}")

            # Infinity minus infinity is zero. This has to be true so that
            # we can for example subtract Multiplier(Bound(0), INF) from
            # Multiplier(Bound(1), INF) to get Multiplier(Bound(1), Bound(1))
            return Bound(0)
        if self.v is None:
            return INF
        try:
            return Bound(self.v - other.v)
        except ValueError as e:
            raise ArithmeticError(*e.args) from e

    def copy(self, /) -> Bound:
        return Bound(self.v)


# Use this for cases where no upper bound is needed
INF = Bound(None)



"""
Because of the circularity between `Pattern`, `Conc` and `Mult`, all three
need to be in the same source file?
"""

from __future__ import annotations

__all__ = (
    "Conc",
    "Mult",
    "Pattern",
    "from_fsm",
)

from dataclasses import dataclass
from enum import Enum, auto
from functools import reduce
from typing import Iterator

from .bound import INF, Bound
from .charclass import NULLCHARCLASS, Charclass
from .fsm import EPSILON, NULL, Fsm, StateType, from_charclass
from .multiplier import ONE, QM, STAR, ZERO, Multiplier


@dataclass(frozen=True)
class Conc:
    """
    A `Conc` (short for "concatenation") is a tuple of `Mult`s i.e. an
    unbroken string of mults occurring one after the other.
    e.g. abcde[^fg]*h{4}[a-z]+(subpattern)(subpattern2)
    To express the empty string, use an empty `Conc`, Conc().
    """

    mults: tuple[Mult, ...]

    def __init__(self, /, *mults: Mult):
        object.__setattr__(self, "mults", tuple(mults))

    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.mults == other.mults

    def __hash__(self, /) -> int:
        return hash(self.mults)

    def __repr__(self, /) -> str:
        args = ", ".join(repr(mult) for mult in self.mults)
        return f"Conc({args})"

    def reduce(self) -> Conc:
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-return-statements

        if self == NULLCONC:
            return self

        if self.empty():
            return NULLCONC

        # Try recursively reducing our mults
        reduced = tuple(mult.reduce() for mult in self.mults)
        if reduced != self.mults:
            return Conc(*reduced).reduce()

        # strip out mults which can only match the empty string
        for i, mult in enumerate(self.mults):
            if (
                # Conc contains "()" (i.e. a `Mult` containing only a `Pattern`
                # containing the empty string)? That can be removed
                # e.g. "a()b" -> "ab"
                mult.multiplicand == Pattern(EMPTYSTRING)
                # If a `Mult` has an empty multiplicand, we can only match it
                # zero times => empty string => remove it entirely
                # e.g. "a[]{0,3}b" -> "ab"
                or (mult.multiplicand.empty() and mult.multiplier.min == Bound(0))
                # Failing that, we have a positive multiplicand which we
                # intend to match zero times. In this case the only possible
                # match is the empty string => remove it
                # e.g. "a[XYZ]{0}b" -> "ab"
                or mult.multiplier == ZERO
            ):
                new = self.mults[:i] + self.mults[i + 1 :]
                return Conc(*new).reduce()

        # We might be able to combine some mults together or at least simplify
        # the multiplier on one of them.
        if len(self.mults) > 1:
            for i in range(len(self.mults) - 1):
                r = self.mults[i]
                s = self.mults[i + 1]

                def to_pattern(multiplicand: Pattern | Charclass, /) -> Pattern:
                    if isinstance(multiplicand, Pattern):
                        return multiplicand
                    return Pattern(Conc(Mult(multiplicand, ONE)))

                # so we can do intersection
                rm_pattern = to_pattern(r.multiplicand)
                sm_pattern = to_pattern(s.multiplicand)
                rm_sm_intersection = None

                # If R = S, then we can squish the multipliers together
                # e.g. ab?b?c -> ab{0,2}c
                if rm_pattern == sm_pattern:
                    squished = Mult(rm_pattern, r.multiplier + s.multiplier)
                    new = self.mults[:i] + (squished,) + self.mults[i + 2 :]
                    return Conc(*new).reduce()

                # If R's language is a subset of S's, then R{a,b}S{c,} reduces
                # to R{a}S{c,}...
                # e.g. \d+\w+ -> \d\w+
                # Do the cheapest checks first
                if r.multiplier.min < r.multiplier.max and s.multiplier.max == INF:
                    rm_sm_intersection = rm_pattern & sm_pattern
                    if rm_sm_intersection.equivalent(rm_pattern):
                        trimmed = Mult(
                            rm_pattern,
                            Multiplier(r.multiplier.min, r.multiplier.min),
                        )
                        new = self.mults[:i] + (trimmed, s) + self.mults[i + 2 :]
                        return Conc(*new).reduce()

                # Conversely, if R is superset of S, then R{c,}S{a,b} reduces
                # to R{c,}S{a}.
                # e.g. [ab]+a? -> [ab]+
                # Do the cheapest checks first
                if r.multiplier.max == INF and s.multiplier.min < s.multiplier.max:
                    if rm_sm_intersection is None:
                        rm_sm_intersection = rm_pattern & sm_pattern
                    if rm_sm_intersection.equivalent(sm_pattern):
                        trimmed = Mult(
                            sm_pattern,
                            Multiplier(s.multiplier.min, s.multiplier.min),
                        )
                        new = self.mults[:i] + (r, trimmed) + self.mults[i + 2 :]
                        return Conc(*new).reduce()

        # Conc contains (among other things) a *singleton* `Mult` containing
        # `Pattern` with only one internal `Conc`? Flatten out.
        # e.g. "a(d(ab|a*c))" -> "ad(ab|a*c)"
        # BUT NOT "a(d(ab|a*c)){2,}"
        # AND NOT "a(d(ab|a*c)|y)"
        for i, mult in enumerate(self.mults):
            if (
                mult.multiplier == ONE
                and isinstance(mult.multiplicand, Pattern)
                and len(mult.multiplicand.concs) == 1
            ):
                (conc,) = mult.multiplicand.concs
                new = self.mults[:i] + conc.mults + self.mults[i + 1 :]
                return Conc(*new).reduce()

        return self

    def to_fsm(self, /) -> Fsm:
        return Fsm.concatenate(EPSILON, *(mult.to_fsm() for mult in self.mults))

    def empty(self, /) -> bool:
        return any(mult.empty() for mult in self.mults)

    def __str__(self, /) -> str:
        return "".join(str(m) for m in self.mults)

    def common(self, other: Conc, /, suffix: bool = False) -> Conc:
        """
        Return the common prefix of these two `Conc`s; that is, the largest
        `Conc` which can be safely beheaded() from the front of both. The
        result could be `EMPTYSTRING`.
        "ZYAA, ZYBB" -> "ZY"
        "CZ, CZ" -> "CZ"
        "YC, ZC" -> ""

        With the "suffix" flag set, works from the end. E.g.:
        "AAZY, BBZY" -> "ZY"
        "CZ, CZ" -> "CZ"
        "CY, CZ" -> ""
        """
        mults = []

        indices = list(range(min(len(self.mults), len(other.mults))))
        # e.g. [0, 1, 2, 3]

        # Work backwards from the end of both `Conc`s instead.
        if suffix:
            indices = [-i - 1 for i in indices]  # e.g. [-1, -2, -3, -4]

        for i in indices:
            x = self.mults[i]
            y = other.mults[i]
            common = x.common(y)

            # Happens when multiplicands disagree (e.g. "A.common(B)") or if
            # the multiplicand is shared but the common multiplier is `ZERO`
            # (e.g. "ABZ*.common(CZ)".)
            if common.multiplier == ZERO:
                break

            mults.append(common)

            # If we did not remove the entirety of both mults, we cannot
            # continue.
            if common != x or common != y:
                break

        if suffix:
            mults = mults[::-1]

        return Conc(*mults)

    def dock(self, other: Conc, /) -> Conc:
        """
        Subtract another `Conc` from this one.
        This is the opposite of concatenation.
        For example, if ABC + DEF = ABCDEF,
        then logically ABCDEF - DEF = ABC.
        """

        # e.g. self has mults at indices [0, 1, 2, 3, 4, 5, 6] len=7
        # e.g. other has mults at indices [0, 1, 2] len=3
        new = list(self.mults)
        for i in reversed(range(len(other.mults))):  # [2, 1, 0]
            # e.g. i = 1, j = 7 - 3 + 1 = 5
            j = len(self.mults) - len(other.mults) + i
            new[j] = new[j].dock(other.mults[i])

            if new[j].multiplier == ZERO:
                # omit that `Mult` entirely since it has been factored out
                del new[j]

            # If the subtraction is incomplete but there is more to
            # other.mults, then we have a problem. For example, "ABC{2} - BC"
            # subtracts the C successfully but leaves something behind,
            # then tries to subtract the B too, which isn't possible
            elif i:
                raise ArithmeticError(f"Can't subtract {other!r} from {self!r}")

        return Conc(*new)

    def behead(self, other: Conc, /) -> Conc:
        """
        As with dock() but the other way around. For example, if
        ABC + DEF = ABCDEF, then ABCDEF.behead(AB) = CDEF.
        """
        # Observe that FEDCBA - BA = FEDC.
        return self.reversed().dock(other.reversed()).reversed()

    def reversed(self, /) -> Conc:
        return Conc(*[mult.reversed() for mult in reversed(self.mults)])


# We need a new state not already used.
class _Outside(Enum):
    """Marker state for use in `from_fsm`."""

    TOKEN = auto()


def from_fsm(f: Fsm) -> Pattern:
    """
    Turn the supplied finite state machine into a `Pattern`. This is
    accomplished using the Brzozowski algebraic method.
    """
    # pylint: disable=too-many-branches

    outside = _Outside.TOKEN

    # The set of strings that would be accepted by this FSM if you started
    # at state i is represented by the regex R_i.
    # If state i has a sole transition "a" to state j, then we know
    # R_i = a R_j.
    # If state i is final, then the empty string is also accepted by this
    # regex.
    # And so on...

    # From this we can build a set of simultaneous equations in len(f.states)
    # variables. This system is easily solved for all variables, but we only
    # need one: R_a, where a is the starting state.

    # The first thing we need to do is organise the states into order of depth,
    # so that when we perform our back-substitutions, we can start with the
    # last (deepest) state and therefore finish with R_a.
    states = [f.initial]
    i = 0
    while i < len(states):
        current = states[i]
        if current in f.map:
            for symbol in sorted(f.map[current]):
                next_state = f.map[current][symbol]
                if next_state not in states:
                    states.append(next_state)
        i += 1

    # Our system of equations is represented like so:
    brz: dict[StateType, dict[StateType | _Outside, Pattern]] = {}

    for a in f.states:
        brz[a] = {}
        for b in f.states:
            brz[a][b] = NULLPATTERN

        if a in f.finals:
            brz[a][outside] = Pattern(EMPTYSTRING)
        else:
            brz[a][outside] = NULLPATTERN

    # Populate it with some initial data.
    for a in f.map:
        for charclass in f.map[a]:
            b = f.map[a][charclass]
            brz[a][b] = Pattern(*brz[a][b].concs, Conc(Mult(charclass, ONE))).reduce()

    # Now perform our back-substitution
    for i in reversed(range(len(states))):
        a = states[i]

        # Before the equation for R_a can be substituted into the other
        # equations, we need to resolve the self-transition (if any).
        # e.g.    R_a = 0 R_a |   1 R_b |   2 R_c
        # becomes R_a =         0*1 R_b | 0*2 R_c
        loop = Mult(brz[a][a], STAR)  # i.e. "0*"
        del brz[a][a]

        for right in brz[a]:
            brz[a][right] = Pattern(Conc(loop, Mult(brz[a][right], ONE))).reduce()

        # Note: even if we're down to our final equation, the above step still
        # needs to be performed before anything is returned.

        # Now we can substitute this equation into all of the previous ones.
        for j in range(i):
            b = states[j]

            # e.g. substituting R_a =  0*1 R_b |      0*2 R_c
            # into              R_b =    3 R_a |        4 R_c | 5 R_d
            # yields            R_b = 30*1 R_b | (30*2|4) R_c | 5 R_d
            univ = brz[b][a]  # i.e. "3"
            del brz[b][a]

            for right in brz[a]:
                brz[b][right] = Pattern(
                    *brz[b][right].concs,
                    Conc(Mult(univ, ONE), Mult(brz[a][right], ONE)),
                ).reduce()

    return brz[f.initial][outside].reduce()


@dataclass(frozen=True)
class Pattern:
    """
    A `Pattern` (also known as an "alt", short for "alternation") is a
    set of `Conc`s. A `Pattern` expresses multiple alternate possibilities.
    When written out as a regex, these would separated by pipes. A
    `Pattern` containing no possibilities is possible and represents a
    regular expression matching no strings whatsoever (there is no
    conventional string form for this).

    e.g. "abc|def(ghi|jkl)" is an alt containing two `Conc`s: "abc" and
    "def(ghi|jkl)". The latter is a `Conc` containing four `Mult`s: "d",
    "e", "f" and "(ghi|jkl)". The latter in turn is a `Mult` consisting of
    an upper bound 1, a lower bound 1, and a multiplicand which is a new
    subpattern, "ghi|jkl". This new subpattern again consists of two
    `Conc`s: "ghi" and "jkl".
    """

    concs: frozenset[Conc]

    def __init__(self, /, *concs: Conc):
        object.__setattr__(self, "concs", frozenset(concs))

    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.concs == other.concs

    def __hash__(self, /) -> int:
        return hash(self.concs)

    def __repr__(self, /) -> str:
        args = ", ".join(repr(conc) for conc in self.concs)
        return f"Pattern({args})"

    def empty(self, /) -> bool:
        return all(conc.empty() for conc in self.concs)

    def intersection(self, other: Pattern, /) -> Pattern:
        combined = self.to_fsm() & other.to_fsm()
        return from_fsm(combined)

    def __and__(self, other: Pattern, /) -> Pattern:
        return self.intersection(other)

    def difference(*elems: Pattern) -> Pattern:
        """
        Return a regular expression which matches any string which `self`
        matches but none of the strings which `other` matches.
        """
        return from_fsm(Fsm.difference(*(elem.to_fsm() for elem in elems)))

    def __sub__(self, other: Pattern, /) -> Pattern:
        return self.difference(other)

    def union(self, other: Pattern, /) -> Pattern:
        return Pattern(*(self.concs | other.concs))

    def __or__(self, other: Pattern, /) -> Pattern:
        return self.union(other)

    def __str__(self, /) -> str:
        if not self.concs:
            raise ValueError(f"Can't serialise {self!r}")
        return "|".join(sorted(str(conc) for conc in self.concs))

    def reduce(self, /) -> Pattern:
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-return-statements

        if self == NULLPATTERN:
            return self

        if self.empty():
            return NULLPATTERN

        # Try recursively reducing our internal `Conc`s.
        reduced = frozenset(c.reduce() for c in self.concs)
        if reduced != self.concs:
            return Pattern(*reduced).reduce()

        # If one of our internal concs is empty, remove it
        for conc in self.concs:
            if conc.empty():
                new = self.concs - {conc}
                return Pattern(*new).reduce()

        # If we have just one `Conc` with just one `Mult` with a multiplier of
        # 1, and the multiplicand is a `Pattern`, pull that up
        if len(self.concs) == 1:
            (conc,) = self.concs
            if (
                len(conc.mults) == 1
                and conc.mults[0].multiplier == ONE
                and isinstance(conc.mults[0].multiplicand, Pattern)
            ):
                return conc.mults[0].multiplicand.reduce()

        # If this `Pattern` contains several `Conc`s each containing just 1
        # `Mult` and their multiplicands agree, we may be able to merge the
        # multipliers.
        # e.g. "a{1,2}|a{3,4}|bc" -> "a{1,4}|bc"
        oldconcs = list(self.concs)  # so we can index the things
        for i, conc1 in enumerate(oldconcs):
            if len(conc1.mults) != 1:
                continue
            multiplicand1 = conc1.mults[0].multiplicand
            for j in range(i + 1, len(oldconcs)):
                conc2 = oldconcs[j]
                if len(conc2.mults) != 1:
                    continue
                multiplicand2 = conc2.mults[0].multiplicand
                if multiplicand2 != multiplicand1:
                    continue
                multiplicand = multiplicand1
                multiplier1 = conc1.mults[0].multiplier
                multiplier2 = conc2.mults[0].multiplier
                if not multiplier1.canunion(multiplier2):
                    continue
                multiplier = multiplier1 | multiplier2
                newconcs = (
                    oldconcs[:i]
                    + oldconcs[i + 1 : j]
                    + oldconcs[j + 1 :]
                    + [Conc(Mult(multiplicand, multiplier))]
                )
                return Pattern(*newconcs).reduce()

        # If this `Pattern` contains several `Conc`s each containing just 1
        # `Mult` each containing just a `Charclass`, with a multiplier of 1,
        # then we can merge those `Charclass`es together.
        # e.g. "0|[1-9]|ab" -> "[0-9]|ab"
        merged_charclass = NULLCHARCLASS
        num_merged = 0
        rest = []
        for conc in self.concs:
            if (
                len(conc.mults) == 1
                and conc.mults[0].multiplier == ONE
                and isinstance(conc.mults[0].multiplicand, Charclass)
            ):
                merged_charclass |= conc.mults[0].multiplicand
                num_merged += 1
            else:
                rest.append(conc)
        if num_merged >= 2:
            rest.append(Conc(Mult(merged_charclass, ONE)))
            return Pattern(*rest).reduce()

        # If one of the present `Pattern`'s `Conc`s is the empty string...
        if EMPTYSTRING in self.concs:
            for conc in self.concs:
                # ...and there is another `Conc`
                # with a single `Mult` whose lower bound is 0...
                if len(conc.mults) == 1 and conc.mults[0].multiplier.min == Bound(0):
                    # Then we can omit the empty string.
                    # E.g. "|(ab)*|def" => "(ab)*|def".
                    return Pattern(*(self.concs - {EMPTYSTRING})).reduce()

            for conc in self.concs:
                # ...and there is another `Conc`
                # with a single `Mult` whose lower bound is 1...
                if len(conc.mults) == 1 and conc.mults[0].multiplier.min == Bound(1):
                    # Then we can merge the empty string into that.
                    # E.g. "|(ab)+|def" => "(ab)*|def".
                    merged_conc = Conc(
                        Mult(conc.mults[0].multiplicand, conc.mults[0].multiplier * QM)
                    )
                    return Pattern(
                        *(self.concs - {EMPTYSTRING, conc} | {merged_conc})
                    ).reduce()

        # If the present `Pattern`'s `Conc`s all have a common prefix, split
        # that out. This increases the depth of the object
        # but it is still arguably simpler/ripe for further reduction
        # e.g. "abc|ade" -> a(bc|de)"
        if len(self.concs) > 1:
            prefix = self._commonconc()
            if prefix != EMPTYSTRING:
                leftovers = self.behead(prefix)
                mults = prefix.mults + (Mult(leftovers, ONE),)
                return Pattern(Conc(*mults)).reduce()

            # Same but for suffixes.
            # e.g. "xyz|stz -> (xy|st)z"
            suffix = self._commonconc(suffix=True)
            if suffix != EMPTYSTRING:
                leftovers = self.dock(suffix)
                mults = (Mult(leftovers, ONE),) + suffix.mults
                return Pattern(Conc(*mults)).reduce()

        return self

    def symmetric_difference(*elems: Pattern) -> Pattern:
        """
        Return a regular expression matching only the strings recognised by
        `self` or `other` but not both.
        """
        return from_fsm(Fsm.symmetric_difference(*(elem.to_fsm() for elem in elems)))

    def __xor__(self, other: Pattern, /) -> Pattern:
        return self.symmetric_difference(other)

    def dock(self, other: Conc, /) -> Pattern:
        """
        The opposite of concatenation. Remove a common suffix from the
        present `Pattern`; that is, from each of its constituent concs.

        AYZ|BYZ|CYZ - YZ -> A|B|C.
        """
        return Pattern(*[conc.dock(other) for conc in self.concs])

    def behead(self, other: Conc, /) -> Pattern:
        """
        Like dock() but the other way around. Remove a common prefix from
        the present `Pattern`; that is, from each of its constituent concs.

        ZA|ZB|ZC.behead(Z) -> A|B|C
        """
        return Pattern(*[conc.behead(other) for conc in self.concs])

    def _commonconc(self, /, suffix: bool = False) -> Conc:
        """
        Find the longest `Conc` which acts as prefix to every `Conc` in
        this `Pattern`. This could be `EMPTYSTRING`. Return the common
        prefix along with all the leftovers after truncating that common
        prefix from each `Conc`.

        "ZA|ZB|ZC" -> "Z", "(A|B|C)"
        "ZA|ZB|ZC|Z" -> "Z", "(A|B|C|)"
        "CZ|CZ" -> "CZ", "()"

        If "suffix" is True, the same result but for suffixes.
        """
        if not self.concs:
            raise ValueError(f"Can't call _commonconc on {self!r}")

        return reduce(lambda x, y: x.common(y, suffix=suffix), self.concs)

    def to_fsm(self, /) -> Fsm:
        return Fsm.union(NULL, *(conc.to_fsm() for conc in self.concs))

    def reversed(self, /) -> Pattern:
        return Pattern(*(c.reversed() for c in self.concs))

    def copy(self, /) -> Pattern:
        """
        For completeness only, since `set.copy()` also exists. `Pattern`s
        are immutable, so I can see only very odd reasons to need this
        """
        return Pattern(*self.concs)

    def equivalent(self, other: Pattern, /) -> bool:
        """
        Two `Pattern`s are equivalent if they recognise the same strings.
        Note that in the general case this is actually quite an intensive
        calculation, but far from unsolvable, as we demonstrate here:
        """
        return self.to_fsm().equivalent(other.to_fsm())

    def times(self, multiplier: Multiplier, /) -> Pattern:
        """
        Equivalent to repeated concatenation. Multiplier consists of a
        minimum and a maximum; maximum may be infinite (for Kleene star
        closure). Call using "a = b * qm"
        """
        return Pattern(Conc(Mult(self, multiplier)))

    def __mul__(self, multiplier: Multiplier, /) -> Pattern:
        return self.times(multiplier)

    def everythingbut(self, /) -> Pattern:
        """
        Return a `Pattern` which will match any string not matched by
        `self`, and which will not match any string matched by `self`.
        Another task which is very difficult in general (and typically
        returns utter garbage when actually printed), but becomes trivial
        to code thanks to FSM routines.
        """
        return from_fsm(self.to_fsm().everythingbut())

    def derive(self, string: str, /) -> Pattern:
        return from_fsm(self.to_fsm().derive(string))

    def isdisjoint(self, other: Pattern, /) -> bool:
        """
        Treat `self` and `other` as sets of strings and see if they are
        disjoint
        """
        return self.to_fsm().isdisjoint(other.to_fsm())

    def matches(self, string: str, /) -> bool:
        return self.to_fsm().accepts(string)

    def __contains__(self, string: str, /) -> bool:
        """
        This lets you use the syntax `"a" in pattern` to see whether the
        string "a" is in the set of strings matched by `pattern`.
        """
        return self.matches(string)

    # pylint: disable=fixme
    # TODO: this is a misuse of __reversed__
    # and should be removed next major version
    def __reversed__(self, /) -> Pattern:
        return self.reversed()

    def cardinality(self, /) -> int:
        """
        Consider the regular expression as a set of strings and return the
        cardinality of that set, or raise an OverflowError if there are
        infinitely many.
        """
        # There is no way to do this other than converting to an FSM, because
        # the `Pattern` may allow duplicate routes, such as "a|a".
        return self.to_fsm().cardinality()

    def __len__(self, /) -> int:
        return self.cardinality()

    def strings(self, /, *, otherchar: str | None = None) -> Iterator[str]:
        """
        Each time next() is called on this iterator, a new string is
        returned which this `Pattern` can match. `StopIteration`
        is raised once all such strings have been returned, although a
        regex with a * in may match infinitely many strings.
        """
        otherchars = [] if otherchar is None else [otherchar]
        return self.to_fsm().strings(otherchars)

    def __iter__(self, /) -> Iterator[str]:
        """
        This allows you to do `for string in pattern` as a list
        comprehension!
        """
        return self.strings()


@dataclass(frozen=True)
class Mult:
    """
    A `Mult` is a combination of a multiplicand with a multiplier (a min
    and a max). The vast majority of characters in regular expressions
    occur without a specific multiplier, which is implicitly equivalent to
    a min of 1 and a max of 1, but many more have explicit multipliers like
    "*" (min = 0, max = INF) and so on.

    e.g. a, b{2}, c?, d*, [efg]{2,5}, f{2,}, (anysubpattern)+, .*, ...
    """

    multiplicand: Charclass | Pattern
    multiplier: Multiplier

    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
            self.multiplicand == other.multiplicand
            and self.multiplier == other.multiplier
        )

    def __hash__(self, /) -> int:
        return hash((self.multiplicand, self.multiplier))

    def __repr__(self, /) -> str:
        return f"Mult({self.multiplicand!r}, {self.multiplier!r})"

    def dock(self, other: Mult, /) -> Mult:
        """
        "Dock" another `Mult` from this one (i.e. remove part of the tail)
        and return the result. The reverse of concatenation. This is a lot
        trickier.
        e.g. a{4,5} - a{3} = a{1,2}
        """
        if other.multiplicand != self.multiplicand:
            raise ArithmeticError(f"Can't subtract {other!r} from {self!r}")
        return Mult(self.multiplicand, self.multiplier - other.multiplier)

    def common(self, other: Mult, /) -> Mult:
        """
        Return the common part of these two mults. This is the largest
        `Mult` which can be safely subtracted from both the originals. The
        multiplier on this `Mult` could be `ZERO`: this is the case if, for
        example, the multiplicands disagree.
        """
        if self.multiplicand == other.multiplicand:
            return Mult(self.multiplicand, self.multiplier.common(other.multiplier))

        # Multiplicands disagree, no common part at all.
        return Mult(NULLCHARCLASS, ZERO)

    def empty(self, /) -> bool:
        return self.multiplicand.empty() and self.multiplier.min > Bound(0)

    def reduce(self, /) -> Mult:
        if self == NULLMULT:
            return self

        # Can't match anything: reduce to empty `Mult`
        if self.empty():
            return NULLMULT

        # Try recursively reducing our multiplicand
        reduced = self.multiplicand.reduce()
        if reduced != self.multiplicand:
            return Mult(reduced, self.multiplier).reduce()

        # If our multiplicand is a `Pattern` containing an empty `Conc`
        # we can pull that "optional" bit out into our own multiplier
        # instead.
        # e.g. (A|B|C|) -> (A|B|C)?
        # e.g. (A|B|C|){2} -> (A|B|C){0,2}
        if (
            isinstance(self.multiplicand, Pattern)
            and EMPTYSTRING in self.multiplicand.concs
            and self.multiplier.canmultiplyby(QM)
        ):
            return Mult(
                Pattern(*(conc for conc in self.multiplicand.concs if conc.mults)),
                self.multiplier * QM,
            ).reduce()

        # If our multiplicand is a `Pattern` containing a single `Conc`
        # containing a single `Mult`, we can scrap the `Pattern` in favour of
        # that `Mult`'s multiplicand
        # e.g. ([ab])* -> [ab]*
        # e.g. ((a))* -> (a)* -> a*
        # NOTE: this logic lives here at the `Mult` level, NOT in
        # `Pattern.reduce` because we want to return another `Mult` (same type)
        if isinstance(self.multiplicand, Pattern) and len(self.multiplicand.concs) == 1:
            (conc,) = self.multiplicand.concs
            if len(conc.mults) == 1 and conc.mults[0].multiplier.canmultiplyby(
                self.multiplier
            ):
                return Mult(
                    conc.mults[0].multiplicand,
                    conc.mults[0].multiplier * self.multiplier,
                ).reduce()

        # no reduction possible
        return self

    def __str__(self, /) -> str:
        if isinstance(self.multiplicand, Pattern):
            return f"({self.multiplicand}){self.multiplier}"
        if isinstance(self.multiplicand, Charclass):
            return f"{self.multiplicand}{self.multiplier}"
        raise TypeError(f"Unknown type {type(self.multiplicand)}")

    def to_fsm(self, /) -> Fsm:
        # worked example: (min, max) = (5, 7) or (5, INF)
        # (mandatory, optional) = (5, 2) or (5, INF)

        unit = (
            from_charclass(self.multiplicand)
            if isinstance(self.multiplicand, Charclass)
            else self.multiplicand.to_fsm()
        )
        # accepts e.g. "ab"

        # Yuck. `mandatory` cannot be infinite: it's just a natural number.
        # However, it uses `Bound`, which describes co-naturals.
        assert self.multiplier.mandatory.v is not None

        # accepts "ababababab"
        mandatory = unit.times(self.multiplier.mandatory.v)

        # unlimited additional copies
        if self.multiplier.optional == INF:
            optional = unit.star()
            # accepts "(ab)*"

        else:
            optional = EPSILON | unit
            # accepts "(ab)?"

            # Implied by `!= INF`.
            assert self.multiplier.optional.v is not None

            optional = optional.times(self.multiplier.optional.v)
            # accepts "(ab)?(ab)?"

        return mandatory.concatenate(optional)

    def reversed(self, /) -> Mult:
        return Mult(self.multiplicand.reversed(), self.multiplier)


NULLMULT = Mult(NULLCHARCLASS, ONE)
NULLCONC = Conc(NULLMULT)
EMPTYSTRING = Conc()
NULLPATTERN = Pattern(NULLCONC)



from __future__ import annotations

import pickle
from copy import copy

import pytest

from .charclass import WORDCHAR, Charclass
from .fsm import EPSILON, NULL, Fsm, from_charclass, unify_alphabets

# pylint: disable=invalid-name,too-many-lines

FixtureA = Fsm

FixtureB = Fsm


def test_addbug() -> None:
    # Odd bug with Fsm.__add__(), exposed by "[bc]*c"
    int5A = Fsm(
        alphabet={Charclass("a"), Charclass("b"), Charclass("c"), ~Charclass("abc")},
        states={0, 1},
        initial=1,
        finals={1},
        map={
            0: {
                ~Charclass("abc"): 0,
                Charclass("a"): 0,
                Charclass("b"): 0,
                Charclass("c"): 0,
            },
            1: {
                ~Charclass("abc"): 0,
                Charclass("a"): 0,
                Charclass("b"): 1,
                Charclass("c"): 1,
            },
        },
    )
    assert int5A.accepts("")

    int5B = Fsm(
        alphabet={Charclass("a"), Charclass("b"), Charclass("c"), ~Charclass("abc")},
        states={0, 1, 2},
        initial=1,
        finals={0},
        map={
            0: {
                ~Charclass("abc"): 2,
                Charclass("a"): 2,
                Charclass("b"): 2,
                Charclass("c"): 2,
            },
            1: {
                ~Charclass("abc"): 2,
                Charclass("a"): 2,
                Charclass("b"): 2,
                Charclass("c"): 0,
            },
            2: {
                ~Charclass("abc"): 2,
                Charclass("a"): 2,
                Charclass("b"): 2,
                Charclass("c"): 2,
            },
        },
    )
    assert int5B.accepts("c")

    int5C = int5A.concatenate(int5B)
    assert int5C.accepts("c")
    # assert int5C.initial == 0


def test_builtins() -> None:
    assert not NULL.accepts("a")
    assert EPSILON.accepts("")
    assert not EPSILON.accepts("a")


@pytest.fixture(name="a")
def fixture_a() -> FixtureA:
    return Fsm(
        alphabet={Charclass("a"), Charclass("b"), ~Charclass("ab")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("a"): 1, Charclass("b"): 2, ~Charclass("ab"): 2},
            1: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
            2: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
        },
    )


def test_a(a: FixtureA) -> None:
    assert not a.accepts("")
    assert a.accepts("a")
    assert not a.accepts("b")


@pytest.fixture(name="b")
def fixture_b() -> FixtureB:
    return Fsm(
        alphabet={Charclass("a"), Charclass("b"), ~Charclass("ab")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("a"): 2, Charclass("b"): 1, ~Charclass("ab"): 2},
            1: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
            2: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
        },
    )


def test_b(b: FixtureB) -> None:
    assert not b.accepts("")
    assert not b.accepts("a")
    assert b.accepts("b")


def test_concatenation_aa(a: FixtureA) -> None:
    concAA = a.concatenate(a)
    assert not concAA.accepts("")
    assert not concAA.accepts("a")
    assert concAA.accepts("aa")
    assert not concAA.accepts("aaa")

    concAA = EPSILON.concatenate(a).concatenate(a)
    assert not concAA.accepts("")
    assert not concAA.accepts("a")
    assert concAA.accepts("aa")
    assert not concAA.accepts("aaa")


def test_concatenation_ab(a: FixtureA, b: FixtureB) -> None:
    concAB = a.concatenate(b)
    assert not concAB.accepts("")
    assert not concAB.accepts("a")
    assert not concAB.accepts("b")
    assert not concAB.accepts("aa")
    assert concAB.accepts("ab")
    assert not concAB.accepts("ba")
    assert not concAB.accepts("bb")


def test_alternation_a(a: FixtureA) -> None:
    altA = a | NULL
    assert not altA.accepts("")
    assert altA.accepts("a")


def test_alternation_ab(a: FixtureA, b: FixtureB) -> None:
    altAB = a | b
    assert not altAB.accepts("")
    assert altAB.accepts("a")
    assert altAB.accepts("b")
    assert not altAB.accepts("aa")
    assert not altAB.accepts("ab")
    assert not altAB.accepts("ba")
    assert not altAB.accepts("bb")


def test_star(a: FixtureA) -> None:
    starA = a.star()
    assert starA.accepts("")
    assert starA.accepts("a")
    assert not starA.accepts("b")
    assert starA.accepts("aaaaaaaaa")


def test_multiply_0(a: FixtureA) -> None:
    zeroA = a.times(0)
    assert zeroA.accepts("")
    assert not zeroA.accepts("a")


def test_multiply_1(a: FixtureA) -> None:
    oneA = a.times(1)
    assert not oneA.accepts("")
    assert oneA.accepts("a")
    assert not oneA.accepts("aa")


def test_multiply_2(a: FixtureA) -> None:
    twoA = a.times(2)
    assert not twoA.accepts("")
    assert not twoA.accepts("a")
    assert twoA.accepts("aa")
    assert not twoA.accepts("aaa")


def test_multiply_7(a: FixtureA) -> None:
    sevenA = a.times(7)
    assert not sevenA.accepts("aaaaaa")
    assert sevenA.accepts("aaaaaaa")
    assert not sevenA.accepts("aaaaaaaa")


def test_optional_mul(a: FixtureA, b: FixtureB) -> None:
    unit = a.concatenate(b)
    # accepts "ab"

    optional = EPSILON | unit
    # accepts "(ab)?
    assert optional.accepts("")
    assert not optional.accepts("a")
    assert not optional.accepts("b")
    assert optional.accepts("ab")
    assert not optional.accepts("aa")

    optional = optional.times(2)
    # accepts "(ab)?(ab)?"
    assert optional.accepts("")
    assert not optional.accepts("a")
    assert not optional.accepts("b")
    assert not optional.accepts("aa")
    assert optional.accepts("ab")
    assert not optional.accepts("ba")
    assert not optional.accepts("bb")
    assert not optional.accepts("aaa")
    assert optional.accepts("abab")


def test_intersection_ab(a: FixtureA, b: FixtureB) -> None:
    intAB = a & b
    assert not intAB.accepts("")
    assert not intAB.accepts("a")
    assert not intAB.accepts("b")


def test_negation(a: FixtureA) -> None:
    everythingbutA = a.everythingbut()
    assert everythingbutA.accepts("")
    assert not everythingbutA.accepts("a")
    assert everythingbutA.accepts("b")
    assert everythingbutA.accepts("aa")
    assert everythingbutA.accepts("ab")


def test_crawl_reduction() -> None:
    # this is "0*1" in heavy disguise. crawl should resolve this duplication
    # Notice how states 2 and 3 behave identically. When resolved together,
    # states 1 and 2&3 also behave identically, so they, too should be resolved
    # (this is impossible to spot before 2 and 3 have been combined).
    merged = Fsm(
        alphabet={Charclass("0"), Charclass("1"), ~Charclass("01")},
        states={1, 2, 3, 4, 5},
        initial=1,
        finals={4},
        map={
            1: {Charclass("0"): 2, Charclass("1"): 4, ~Charclass("01"): 5},
            2: {Charclass("0"): 3, Charclass("1"): 4, ~Charclass("01"): 5},
            3: {Charclass("0"): 3, Charclass("1"): 4, ~Charclass("01"): 5},
            4: {Charclass("0"): 5, Charclass("1"): 5, ~Charclass("01"): 5},
            5: {Charclass("0"): 5, Charclass("1"): 5, ~Charclass("01"): 5},
        },
    ).reduce()
    assert len(merged.states) == 3


def test_bug_28() -> None:
    # This is (ab*)* and it caused some defects.
    abstar = Fsm(
        alphabet={Charclass("a"), Charclass("b"), ~Charclass("ab")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("a"): 1, Charclass("b"): 2, ~Charclass("ab"): 2},
            1: {Charclass("a"): 2, Charclass("b"): 1, ~Charclass("ab"): 2},
            2: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
        },
    )
    assert abstar.accepts("a")
    assert not abstar.accepts("b")
    assert abstar.accepts("ab")
    assert abstar.accepts("abb")
    abstarstar = abstar.star()
    assert abstarstar.accepts("a")
    assert not abstarstar.accepts("b")
    assert abstarstar.accepts("ab")
    assert not abstar.star().accepts("bb")


def test_star_advanced() -> None:
    # This is (a*ba)*. Naively connecting the final states to the initial state
    # gives the incorrect result here.
    starred = Fsm(
        alphabet={Charclass("a"), Charclass("b"), ~Charclass("ab")},
        states={0, 1, 2, 3},
        initial=0,
        finals={2},
        map={
            0: {Charclass("a"): 0, Charclass("b"): 1, ~Charclass("ab"): 3},
            1: {Charclass("a"): 2, Charclass("b"): 3, ~Charclass("ab"): 3},
            2: {Charclass("a"): 3, Charclass("b"): 3, ~Charclass("ab"): 3},
            3: {Charclass("a"): 3, Charclass("b"): 3, ~Charclass("ab"): 3},
        },
    ).star()
    assert starred.alphabet == frozenset(
        [
            Charclass("a"),
            Charclass("b"),
            ~Charclass("ab"),
        ]
    )
    assert starred.accepts("")
    assert not starred.accepts("a")
    assert not starred.accepts("b")
    assert not starred.accepts("aa")
    assert starred.accepts("ba")
    assert starred.accepts("aba")
    assert starred.accepts("aaba")
    assert not starred.accepts("aabb")
    assert starred.accepts("abababa")


def test_reduce() -> None:
    # FSM accepts no strings but has 3 states, needs only 1
    asdf = Fsm(
        alphabet={Charclass("x"), ~Charclass("x")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("x"): 2, ~Charclass("x"): 2},
            1: {Charclass("x"): 2, ~Charclass("x"): 2},
            2: {Charclass("x"): 2, ~Charclass("x"): 2},
        },
    )
    asdf = asdf.reduce()
    assert len(asdf.states) == 1


def test_reverse_abc() -> None:
    abc = Fsm(
        alphabet={Charclass("a"), Charclass("b"), Charclass("c"), ~Charclass("abc")},
        states={0, 1, 2, 3, 4},
        initial=0,
        finals={3},
        map={
            0: {
                Charclass("a"): 1,
                Charclass("b"): 4,
                Charclass("c"): 4,
                ~Charclass("abc"): 4,
            },
            1: {
                Charclass("a"): 4,
                Charclass("b"): 2,
                Charclass("c"): 4,
                ~Charclass("abc"): 4,
            },
            2: {
                Charclass("a"): 4,
                Charclass("b"): 4,
                Charclass("c"): 3,
                ~Charclass("abc"): 4,
            },
            3: {
                Charclass("a"): 4,
                Charclass("b"): 4,
                Charclass("c"): 4,
                ~Charclass("abc"): 4,
            },
            4: {
                Charclass("a"): 4,
                Charclass("b"): 4,
                Charclass("c"): 4,
                ~Charclass("abc"): 4,
            },
        },
    )
    cba = abc.reversed()
    assert cba.accepts("cba")


def test_reverse_brzozowski() -> None:
    # This is (a|b)*a(a|b)
    brzozowski = Fsm(
        alphabet={Charclass("a"), Charclass("b"), ~Charclass("ab")},
        states={0, 1, 2, 3, 4, 5},
        initial=0,
        finals={2, 4},
        map={
            0: {Charclass("a"): 1, Charclass("b"): 3, ~Charclass("ab"): 5},
            1: {Charclass("a"): 2, Charclass("b"): 4, ~Charclass("ab"): 5},
            2: {Charclass("a"): 2, Charclass("b"): 4, ~Charclass("ab"): 5},
            3: {Charclass("a"): 1, Charclass("b"): 3, ~Charclass("ab"): 5},
            4: {Charclass("a"): 1, Charclass("b"): 3, ~Charclass("ab"): 5},
            5: {Charclass("a"): 5, Charclass("b"): 5, ~Charclass("ab"): 5},
        },
    )
    assert brzozowski.accepts("aa")
    assert brzozowski.accepts("ab")
    assert brzozowski.accepts("aab")
    assert brzozowski.accepts("bab")
    assert brzozowski.accepts("abbbbbbbab")
    assert not brzozowski.accepts("")
    assert not brzozowski.accepts("a")
    assert not brzozowski.accepts("b")
    assert not brzozowski.accepts("ba")
    assert not brzozowski.accepts("bb")
    assert not brzozowski.accepts("bbbbbbbbbbbb")

    # So this is (a|b)a(a|b)*
    b2 = brzozowski.reversed()
    assert b2.accepts("aa")
    assert b2.accepts("ba")
    assert b2.accepts("baa")
    assert b2.accepts("bab")
    assert b2.accepts("babbbbbbba")
    assert not b2.accepts("")
    assert not b2.accepts("a")
    assert not b2.accepts("b")
    assert not b2.accepts("ab")
    assert not b2.accepts("bb")
    assert not b2.accepts("bbbbbbbbbbbb")

    # Test string generator functionality.
    gen = b2.strings([])
    assert next(gen) == "aa"
    assert next(gen) == "ba"
    assert next(gen) == "aaa"
    assert next(gen) == "aab"
    assert next(gen) == "baa"
    assert next(gen) == "bab"
    assert next(gen) == "aaaa"


def test_reverse_epsilon() -> None:
    # EPSILON reversed is EPSILON
    assert EPSILON.reversed().accepts("")


def test_binary_3() -> None:
    # Binary numbers divisible by 3.
    # Disallows the empty string
    # Allows "0" on its own, but not leading zeroes.
    div3 = Fsm(
        alphabet={Charclass("0"), Charclass("1"), ~Charclass("01")},
        states={-2, -1, 0, 1, 2, 3},
        initial=-2,
        finals={-1, 0},
        map={
            -2: {Charclass("0"): -1, Charclass("1"): 1, ~Charclass("01"): 3},
            -1: {Charclass("0"): 3, Charclass("1"): 3, ~Charclass("01"): 3},
            0: {Charclass("0"): 0, Charclass("1"): 1, ~Charclass("01"): 3},
            1: {Charclass("0"): 2, Charclass("1"): 0, ~Charclass("01"): 3},
            2: {Charclass("0"): 1, Charclass("1"): 2, ~Charclass("01"): 3},
            3: {Charclass("0"): 3, Charclass("1"): 3, ~Charclass("01"): 3},
        },
    )
    assert not div3.accepts("")
    assert div3.accepts("0")
    assert not div3.accepts("1")
    assert not div3.accepts("00")
    assert not div3.accepts("01")
    assert not div3.accepts("10")
    assert div3.accepts("11")
    assert not div3.accepts("000")
    assert not div3.accepts("001")
    assert not div3.accepts("010")
    assert not div3.accepts("011")
    assert not div3.accepts("100")
    assert not div3.accepts("101")
    assert div3.accepts("110")
    assert not div3.accepts("111")
    assert not div3.accepts("0000")
    assert not div3.accepts("0001")
    assert not div3.accepts("0010")
    assert not div3.accepts("0011")
    assert not div3.accepts("0100")
    assert not div3.accepts("0101")
    assert not div3.accepts("0110")
    assert not div3.accepts("0111")
    assert not div3.accepts("1000")
    assert div3.accepts("1001")


def test_invalid_fsms() -> None:
    # initial state 1 is not a state
    with pytest.raises(ValueError, match="Initial state"):
        Fsm(alphabet={}, states={}, initial=1, finals=(), map={})

    # final state 2 not a state
    with pytest.raises(ValueError, match="Final states"):
        Fsm(alphabet={}, states={1}, initial=1, finals={2}, map={})

    # invalid transition for state 1, symbol "a"
    with pytest.raises(ValueError, match="Transition.+leads to.+not a state"):
        Fsm(
            alphabet={Charclass("a")},
            states={1},
            initial=1,
            finals=(),
            map={1: {Charclass("a"): 2}},
        )

    # invalid transition from unknown state
    with pytest.raises(ValueError, match="Transition.+unknown state"):
        Fsm(
            alphabet={Charclass("a")},
            states={1, 2},
            initial=1,
            finals=(),
            map={3: {Charclass("a"): 2}},
        )

    # invalid transition table includes symbol outside of alphabet
    with pytest.raises(ValueError, match="Invalid symbol"):
        Fsm(
            alphabet={Charclass("a")},
            states={1, 2},
            initial=1,
            finals=(),
            map={1: {Charclass("a"): 2, Charclass("b"): 2}},
        )


def test_bad_multiplier(a: FixtureA) -> None:
    with pytest.raises(ArithmeticError, match="Can't multiply"):
        _ = a.times(-1)


def test_anything_else_acceptance() -> None:
    a = Fsm(
        alphabet={Charclass("a"), Charclass("b"), Charclass("c"), ~Charclass("abc")},
        states={1},
        initial=1,
        finals={1},
        map={
            1: {
                Charclass("a"): 1,
                Charclass("b"): 1,
                Charclass("c"): 1,
                ~Charclass("abc"): 1,
            }
        },
    )
    assert a.accepts("d")


def test_difference(a: FixtureA, b: FixtureB) -> None:
    aorb = Fsm(
        alphabet={Charclass("a"), Charclass("b"), ~Charclass("ab")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("a"): 1, Charclass("b"): 1, ~Charclass("ab"): 2},
            1: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
            2: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
        },
    )

    assert not list((a ^ a).strings([]))
    assert not list((b ^ b).strings([]))
    assert list((a ^ b).strings([])) == ["a", "b"]
    assert list((aorb ^ a).strings([])) == ["b"]


def test_empty(a: FixtureA, b: FixtureB) -> None:
    assert not a.empty()
    assert not b.empty()

    assert Fsm(
        alphabet={~Charclass()},
        states={0, 1},
        initial=0,
        finals={1},
        map={0: {~Charclass(): 0}, 1: {~Charclass(): 0}},
    ).empty()

    assert not Fsm(
        alphabet={~Charclass()},
        states={0},
        initial=0,
        finals={0},
        map={0: {~Charclass(): 0}},
    ).empty()

    assert Fsm(
        alphabet={Charclass("a"), Charclass("b"), ~Charclass("ab")},
        states={0, 1, 2, 3},
        initial=0,
        finals={3},
        map={
            0: {Charclass("a"): 1, Charclass("b"): 1, ~Charclass("ab"): 2},
            1: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
            2: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
            3: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
        },
    ).empty()


def test_equivalent(a: FixtureA, b: FixtureB) -> None:
    assert (a | b).equivalent(b | a)


def test_eq_ne(a: FixtureA, b: FixtureB) -> None:
    # pylint: disable=comparison-with-itself

    assert a == a
    assert b == b
    assert a != b
    assert b != a
    assert (a | b) == (b | a)


@pytest.mark.parametrize(
    argnames="other",
    argvalues=(
        17,
        (14,),
        "blenny",
        "a",
        ("a",),
    ),
)
def test_eq_ne_het(a: FixtureA, other: object) -> None:
    # eq
    assert not a == other
    # eq, symmetric
    assert not other == a

    # neq
    assert a != other
    # neq, symmetric
    assert other != a


def test_dead_default() -> None:
    """
    Old test from when you used to be able to have sparse maps
    """
    blockquote = Fsm(
        alphabet={Charclass("/"), Charclass("*"), ~Charclass("/*")},
        states={0, 1, 2, 3, 4, 5},
        initial=0,
        finals={4},
        map={
            0: {Charclass("/"): 1, ~Charclass("/*"): 5, Charclass("*"): 5},
            1: {Charclass("/"): 5, ~Charclass("/*"): 5, Charclass("*"): 2},
            2: {Charclass("/"): 2, ~Charclass("/*"): 2, Charclass("*"): 3},
            3: {Charclass("/"): 4, ~Charclass("/*"): 2, Charclass("*"): 3},
            4: {Charclass("/"): 5, ~Charclass("/*"): 5, Charclass("*"): 5},
            5: {Charclass("/"): 5, ~Charclass("/*"): 5, Charclass("*"): 5},
        },
    )
    assert blockquote.accepts("/*whatever*/")
    assert not blockquote.accepts("**whatever*/")
    assert (
        str(blockquote)
        == "  name final? \\* / [^*/] \n"
        + "-------------------------\n"
        + "* 0    False  5  1 5     \n"
        + "  1    False  2  5 5     \n"
        + "  2    False  3  2 2     \n"
        + "  3    False  3  4 2     \n"
        + "  4    True   5  5 5     \n"
        + "  5    False  5  5 5     \n"
    )
    _ = blockquote | blockquote
    _ = blockquote & blockquote
    _ = blockquote ^ blockquote
    # Fsm does not support the `Reversible` protocol, because its
    # `__reversed__` implementation does not return an iterator.
    # Even if it did, it would not conform semantically because it returns an
    # iterable of the reversed strings, not a reversed iteration of those
    # strings.
    # reversed(blockquote)
    blockquote.reversed()
    assert not blockquote.everythingbut().accepts("/*whatever*/")

    # deliberately seek oblivion
    assert blockquote.everythingbut().accepts("*")

    assert blockquote.islive(3)
    assert blockquote.islive(4)
    assert not blockquote.islive(5)
    gen = blockquote.strings([])
    assert next(gen) == "/**/"


def test_alphabet_unions() -> None:
    # It should now be possible to compute the union of
    # FSMs with disagreeing alphabets!
    a = Fsm(
        alphabet={Charclass("a"), ~Charclass("a")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("a"): 1, ~Charclass("a"): 2},
            1: {Charclass("a"): 1, ~Charclass("a"): 2},
            2: {Charclass("a"): 2, ~Charclass("a"): 2},
        },
    )

    b = Fsm(
        alphabet={Charclass("b"), ~Charclass("b")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("b"): 1, ~Charclass("b"): 2},
            1: {Charclass("b"): 1, ~Charclass("b"): 2},
            2: {Charclass("b"): 2, ~Charclass("b"): 2},
        },
    )

    assert (a | b).accepts("a")
    assert (a | b).accepts("b")
    assert (a & b).empty()
    assert a.concatenate(b).accepts("ab")
    assert (a ^ b).accepts("a")
    assert (a ^ b).accepts("b")


def test_new_set_methods(a: FixtureA, b: FixtureB) -> None:
    # A whole bunch of new methods were added to the FSM module to enable FSMs
    # to function exactly as if they were sets of strings (symbol lists), see:
    # https://docs.python.org/3/library/stdtypes.html#set-types-set-frozenset
    # But do they work?
    assert len(a) == 1
    assert len((a | b).times(4)) == 16

    with pytest.raises(OverflowError):
        len(a.star())

    # "in"
    assert "a" in a
    assert "a" not in b

    # List comprehension!
    four = (a | b).times(2)
    for string in four:
        assert string == "aa"
        break
    assert tuple(four) == (
        "aa",
        "ab",
        "ba",
        "bb",
    )

    # set.union() imitation
    assert Fsm.union(a, b) == a.union(b)
    # pylint: disable-next=compare-to-zero
    assert len(Fsm.union()) == 0
    assert Fsm.intersection(a, b) == a.intersection(b)

    # This takes a little explaining. In general, `a & b & c` is equivalent to
    # `EVERYTHING & a & b & c` where `EVERYTHING` is an FSM accepting every
    # possible string. Similarly `a` is equivalent to `EVERYTHING & a`, and the
    # intersection of no sets at all is... `EVERYTHING`.
    int_none = Fsm.intersection()
    with pytest.raises(OverflowError):
        len(int_none)
    assert "" in int_none

    assert (a | b).difference(a) == Fsm.difference((a | b), a) == (a | b) - a == b
    assert (
        (a | b).difference(a, b)
        == Fsm.difference((a | b), a, b)
        == (a | b) - a - b
        == NULL
    )
    assert a.symmetric_difference(b) == Fsm.symmetric_difference(a, b) == a ^ b
    assert a.isdisjoint(b)
    assert a <= (a | b)
    assert a < (a | b)
    assert a != (a | b)
    assert (a | b) > a
    assert (a | b) >= a

    assert list(a.concatenate(a, a).strings([])) == ["aaa"]
    assert list(a.concatenate().strings([])) == ["a"]
    assert list(Fsm.concatenate(b, a, b).strings([])) == ["bab"]
    assert not list(Fsm.concatenate().strings([]))


def test_copy(a: FixtureA) -> None:
    # fsm.copy() and frozenset().copy() both preserve identity, because they
    # are immutable. This is just showing that we give the same behaviour.
    copyables: tuple[Fsm | frozenset[str], ...] = (a, frozenset("abc"))
    for x in copyables:
        assert x.copy() is x

    # Same, via the `__copy__` method.
    for x in copyables:
        assert copy(x) is x


def test_oblivion_crawl() -> None:
    # Old test from when we used to have a suppressed/secret "oblivion state"
    abc = Fsm(
        alphabet={Charclass("a"), Charclass("b"), Charclass("c"), ~Charclass("abc")},
        states={0, 1, 2, 3, 4},
        initial=0,
        finals={3},
        map={
            0: {
                Charclass("a"): 1,
                Charclass("b"): 2,
                Charclass("c"): 4,
                ~Charclass("abc"): 4,
            },
            1: {
                Charclass("a"): 4,
                Charclass("b"): 2,
                Charclass("c"): 4,
                ~Charclass("abc"): 4,
            },
            2: {
                Charclass("a"): 4,
                Charclass("b"): 4,
                Charclass("c"): 3,
                ~Charclass("abc"): 4,
            },
            3: {
                Charclass("a"): 4,
                Charclass("b"): 4,
                Charclass("c"): 4,
                ~Charclass("abc"): 4,
            },
            4: {
                Charclass("a"): 4,
                Charclass("b"): 4,
                Charclass("c"): 4,
                ~Charclass("abc"): 4,
            },
        },
    )
    assert len(Fsm.concatenate(abc, abc).states) == 8
    assert len(abc.star().states) == 4
    assert len(abc.times(3).states) == 11
    assert len(abc.reversed().states) == 5
    assert len((abc | abc).states) == 5
    assert len((abc & abc).states) == 5
    assert len((abc ^ abc).states) == 1
    assert len((abc - abc).states) == 1


def test_concatenate_bug(a: FixtureA) -> None:
    # This exposes a defect in Fsm.concatenate.
    assert Fsm.concatenate(a, EPSILON, a).accepts("aa")
    assert Fsm.concatenate(
        a,
        EPSILON,
        EPSILON,
        a,
    ).accepts("aa")


def test_derive(a: FixtureA) -> None:
    # Just some basic tests because this is mainly a regex thing.
    assert a.derive("a") == EPSILON
    assert a.derive("b") == NULL

    assert a.times(3).derive("a") == a.times(2)
    assert (a.star() - EPSILON).derive("a") == a.star()


def test_bug_36() -> None:
    # This is /.*/
    etc1 = Fsm(
        alphabet={~Charclass()},
        states={0},
        initial=0,
        finals={0},
        map={
            0: {~Charclass(): 0},
        },
    )

    # This is /s.*/
    etc2 = Fsm(
        alphabet={Charclass("s"), ~Charclass("s")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("s"): 1, ~Charclass("s"): 2},
            1: {Charclass("s"): 1, ~Charclass("s"): 1},
            2: {Charclass("s"): 2, ~Charclass("s"): 2},
        },
    )

    both = etc1 & etc2
    assert etc1.accepts("")
    assert etc1.accepts("s")
    assert etc1.accepts("ts")
    assert not etc2.accepts("")
    assert etc2.accepts("s")
    assert not etc2.accepts("ts")
    assert both.alphabet == {~Charclass("s"), Charclass("s")}
    assert both.accepts("s")


def test_add_anything_else() -> None:
    # [^a]
    fsm1 = Fsm(
        alphabet={Charclass("a"), ~Charclass("a")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("a"): 2, ~Charclass("a"): 1},
            1: {Charclass("a"): 2, ~Charclass("a"): 1},
            2: {Charclass("a"): 2, ~Charclass("a"): 2},
        },
    )

    # [^b]
    fsm2 = Fsm(
        alphabet={Charclass("b"), ~Charclass("b")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("b"): 2, ~Charclass("b"): 1},
            1: {Charclass("b"): 2, ~Charclass("b"): 1},
            2: {Charclass("b"): 2, ~Charclass("b"): 2},
        },
    )
    assert fsm1.concatenate(fsm2).accepts("ba")


def test_anything_else_pickle() -> None:
    # [^z]*
    fsm1 = Fsm(
        alphabet={Charclass("z"), ~Charclass("z")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("z"): 2, ~Charclass("z"): 1},
            1: {Charclass("z"): 2, ~Charclass("z"): 1},
            2: {Charclass("z"): 2, ~Charclass("z"): 2},
        },
    )

    fsm1_unpickled = pickle.loads(pickle.dumps(fsm1))

    # Newly-created instance.
    assert fsm1_unpickled is not fsm1

    # but equivalent.
    assert fsm1 == fsm1_unpickled

    assert fsm1_unpickled.alphabet == {Charclass("z"), ~Charclass("z")}


def test_replace_alphabet() -> None:
    # [^z]*
    fsm1 = Fsm(
        alphabet={Charclass("z"), ~Charclass("z")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("z"): 2, ~Charclass("z"): 1},
            1: {Charclass("z"): 2, ~Charclass("z"): 1},
            2: {Charclass("z"): 2, ~Charclass("z"): 2},
        },
    )

    fsm2 = fsm1.replace_alphabet(
        {
            Charclass("z"): [Charclass("a"), Charclass("b")],
            ~Charclass("z"): [Charclass("c"), ~Charclass("abc")],
        }
    )

    assert fsm2.map == {
        0: {
            Charclass("a"): 2,
            Charclass("b"): 2,
            Charclass("c"): 1,
            ~Charclass("abc"): 1,
        },
        1: {
            Charclass("a"): 2,
            Charclass("b"): 2,
            Charclass("c"): 1,
            ~Charclass("abc"): 1,
        },
        2: {
            Charclass("a"): 2,
            Charclass("b"): 2,
            Charclass("c"): 2,
            ~Charclass("abc"): 2,
        },
    }


def test_replace_alphabet_2() -> None:
    # [^z]*
    fsm1 = Fsm(
        alphabet={Charclass("z"), ~Charclass("z")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("z"): 2, ~Charclass("z"): 1},
            1: {Charclass("z"): 2, ~Charclass("z"): 1},
            2: {Charclass("z"): 2, ~Charclass("z"): 2},
        },
    )

    fsm2 = fsm1.replace_alphabet({Charclass("z"): [~Charclass()], ~Charclass("z"): []})

    assert fsm2.map == {
        0: {~Charclass(): 2},
        1: {~Charclass(): 2},
        2: {~Charclass(): 2},
    }


def test_charclass_fsm() -> None:
    # "[^a]"
    nota = from_charclass(~Charclass("a"))
    assert nota.alphabet == {Charclass("a"), ~Charclass("a")}
    assert nota.accepts("b")
    assert nota.accepts("c")


def test_charclass_fsm_2() -> None:
    bc = from_charclass(Charclass("bc"))
    assert bc.alphabet == {Charclass("bc"), ~Charclass("bc")}
    assert bc.map == {
        0: {Charclass("bc"): 1, ~Charclass("bc"): 2},
        1: {Charclass("bc"): 2, ~Charclass("bc"): 2},
        2: {Charclass("bc"): 2, ~Charclass("bc"): 2},
    }
    assert not bc.accepts("")
    assert not bc.accepts("a")
    assert bc.accepts("b")
    assert bc.accepts("c")
    assert not bc.accepts("d")
    assert not bc.accepts("bc")


def test_charclass_fsm_3() -> None:
    notbc = from_charclass(~Charclass("bc"))
    assert notbc.alphabet == {Charclass("bc"), ~Charclass("bc")}
    assert notbc.map == {
        0: {Charclass("bc"): 2, ~Charclass("bc"): 1},
        1: {Charclass("bc"): 2, ~Charclass("bc"): 2},
        2: {Charclass("bc"): 2, ~Charclass("bc"): 2},
    }
    assert not notbc.accepts("")
    assert notbc.accepts("a")
    assert not notbc.accepts("b")
    assert not notbc.accepts("c")
    assert notbc.accepts("d")
    assert not notbc.accepts("aa")


def test_charclass_fsm_bad() -> None:
    wordchar = from_charclass(WORDCHAR)
    assert len(wordchar.alphabet) == 2
    assert len(wordchar.map[0].values()) == 2


def test_unify_alphabets() -> None:
    a = Fsm(
        alphabet={Charclass("a"), ~Charclass("a")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("a"): 1, ~Charclass("a"): 2},
            1: {Charclass("a"): 2, ~Charclass("a"): 2},
            2: {Charclass("a"): 2, ~Charclass("a"): 2},
        },
    )
    assert a.alphabet == {Charclass("a"), ~Charclass("a")}

    b = Fsm(
        alphabet={Charclass("b"), ~Charclass("b")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("b"): 1, ~Charclass("b"): 2},
            1: {Charclass("b"): 2, ~Charclass("b"): 2},
            2: {Charclass("b"): 2, ~Charclass("b"): 2},
        },
    )
    assert b.alphabet == {Charclass("b"), ~Charclass("b")}

    [a2, b2] = unify_alphabets((a, b))
    assert a2.alphabet == {Charclass("a"), Charclass("b"), ~Charclass("ab")}
    assert a2.map == {
        0: {Charclass("a"): 1, Charclass("b"): 2, ~Charclass("ab"): 2},
        1: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
        2: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
    }
    assert b2.alphabet == {Charclass("a"), Charclass("b"), ~Charclass("ab")}
    assert b2.map == {
        0: {Charclass("a"): 2, Charclass("b"): 1, ~Charclass("ab"): 2},
        1: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
        2: {Charclass("a"): 2, Charclass("b"): 2, ~Charclass("ab"): 2},
    }


def test_bad_alphabets() -> None:
    with pytest.raises(ValueError, match="has overlaps"):
        Fsm(
            alphabet={Charclass("a"), Charclass("ab")},
            states={0},
            initial=0,
            finals=(),
            map={0: {Charclass("a"): 0, Charclass("ab"): 0}},
        )

    with pytest.raises(ValueError, match="not a proper partition"):
        Fsm(
            alphabet={Charclass("a")},
            states={0},
            initial=0,
            finals=(),
            map={0: {Charclass("a"): 0}},
        )

    with pytest.raises(ValueError, match="not a proper partition"):
        Fsm(
            alphabet={~Charclass("b")},
            states={0},
            initial=0,
            finals=(),
            map={0: {~Charclass("b"): 0}},
        )

    with pytest.raises(ValueError, match="not a proper partition"):
        Fsm(
            alphabet={Charclass("a"), ~Charclass("ab")},
            states={0},
            initial=0,
            finals=(),
            map={0: {Charclass("a"): 0, ~Charclass("ab"): 0}},
        )


def test_larger_charclasses() -> None:
    aorb = Fsm(
        alphabet={Charclass("ab"), ~Charclass("ab")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("ab"): 1, ~Charclass("ab"): 2},
            1: {Charclass("ab"): 2, ~Charclass("ab"): 2},
            2: {Charclass("ab"): 2, ~Charclass("ab"): 2},
        },
    )
    assert not aorb.accepts("")
    assert aorb.accepts("a")
    assert aorb.accepts("b")
    assert not aorb.accepts("c")
    assert not aorb.accepts("aa")


def test_nightmare_charclass() -> None:
    # This consumes over a million different possible characters
    # Previously this would bring the package to its knees, not anymore!
    nightmare = from_charclass(
        Charclass(
            (
                ("\t", "\t"),
                ("\n", "\n"),
                ("\r", "\r"),
                (" ", "\uD7FF"),
                ("\uE000", "\uFFFD"),
                ("\U00010000", "\U0010FFFF"),
            )
        )
    )
    assert nightmare.accepts("\uE123")



from __future__ import annotations

import unicodedata

from .charclass import (
    DIGIT,
    DOT,
    NONDIGITCHAR,
    NONSPACECHAR,
    NONWORDCHAR,
    NULLCHARCLASS,
    SPACECHAR,
    WORDCHAR,
    Charclass,
    collapse_ord_ranges,
    repartition,
)


def test_collapse_ord_ranges_0() -> None:
    assert collapse_ord_ranges([(1, 2)]) == [(1, 2)]


def test_collapse_ord_ranges_1a() -> None:
    assert collapse_ord_ranges(
        [(1, 1), (3, 4), (10, 11), (13, 17), (7, 7)],
    ) == [(1, 1), (3, 4), (7, 7), (10, 11), (13, 17)]


def test_collapse_ord_ranges_1b() -> None:
    assert collapse_ord_ranges([(5, 16), (1, 1)]) == [(1, 1), (5, 16)]
    assert collapse_ord_ranges([(5, 16), (1, 2)]) == [(1, 2), (5, 16)]
    assert collapse_ord_ranges([(5, 16), (1, 3)]) == [(1, 3), (5, 16)]
    assert collapse_ord_ranges([(5, 16), (1, 4)]) == [(1, 16)]
    assert collapse_ord_ranges([(5, 16), (1, 5)]) == [(1, 16)]
    assert collapse_ord_ranges([(5, 16), (1, 16)]) == [(1, 16)]
    assert collapse_ord_ranges([(5, 16), (1, 17)]) == [(1, 17)]
    assert collapse_ord_ranges([(5, 16), (1, 18)]) == [(1, 18)]
    assert collapse_ord_ranges([(5, 16), (4, 4)]) == [(4, 16)]
    assert collapse_ord_ranges([(5, 16), (5, 5)]) == [(5, 16)]
    assert collapse_ord_ranges([(5, 16), (5, 18)]) == [(5, 18)]
    assert collapse_ord_ranges([(5, 16), (7, 8)]) == [(5, 16)]
    assert collapse_ord_ranges([(5, 16), (10, 20)]) == [(5, 20)]
    assert collapse_ord_ranges([(5, 16), (16, 20)]) == [(5, 20)]
    assert collapse_ord_ranges([(5, 16), (17, 20)]) == [(5, 20)]
    assert collapse_ord_ranges([(5, 16), (18, 20)]) == [(5, 16), (18, 20)]


def test_collapse_ord_ranges_2() -> None:
    assert collapse_ord_ranges([(1, 2), (11, 12), (5, 6)]) == [(1, 2), (5, 6), (11, 12)]
    assert collapse_ord_ranges([(1, 2), (11, 12), (3, 6)]) == [(1, 6), (11, 12)]
    assert collapse_ord_ranges([(1, 2), (11, 12), (2, 6)]) == [(1, 6), (11, 12)]
    assert collapse_ord_ranges([(1, 2), (11, 12), (5, 9)]) == [(1, 2), (5, 9), (11, 12)]
    assert collapse_ord_ranges([(1, 2), (11, 12), (5, 10)]) == [(1, 2), (5, 12)]
    assert collapse_ord_ranges([(1, 2), (11, 12), (-2, -1)]) == [
        (-2, -1),
        (1, 2),
        (11, 12),
    ]
    assert collapse_ord_ranges([(1, 2), (11, 12), (0, 20)]) == [(0, 20)]


def test_charclass_equality() -> None:
    assert Charclass("a") == Charclass("a")
    assert ~Charclass("a") == ~Charclass("a")
    assert ~Charclass("a") != Charclass("a")
    assert Charclass("ab") == Charclass("ba")


def test_charclass_ctor() -> None:
    assert not Charclass("ab").negated
    assert not Charclass("ab", negated=False).negated
    assert Charclass("ab", negated=True).negated


def test_repr() -> None:
    assert repr(~Charclass("a")) == "~Charclass((('a', 'a'),))"


def test_issubset() -> None:
    assert Charclass("a").issubset(Charclass("a"))
    assert not Charclass("a").issubset(Charclass("b"))
    assert Charclass("a").issubset(Charclass((("a", "b"),)))
    assert Charclass("a").issubset(~Charclass("b"))
    assert not (~Charclass("a")).issubset(Charclass("b"))
    assert (~Charclass("a")).issubset(DOT)


def test_charclass_str() -> None:
    assert str(WORDCHAR) == "\\w"
    assert str(DIGIT) == "\\d"
    assert str(SPACECHAR) == "\\s"
    assert str(Charclass("a")) == "a"
    assert str(Charclass("{")) == "\\{"
    assert str(Charclass("\t")) == "\\t"
    assert str(Charclass("ab")) == "[ab]"
    assert str(Charclass("a{")) == "[a{]"
    assert str(Charclass("a\t")) == "[\\ta]"
    assert str(Charclass("a-")) == "[\\-a]"
    assert str(Charclass("a[")) == "[\\[a]"
    assert str(Charclass("a]")) == "[\\]a]"
    assert str(Charclass("ab")) == "[ab]"
    assert str(Charclass("abc")) == "[abc]"
    assert str(Charclass("abcd")) == "[a-d]"
    assert str(Charclass("abcdfghi")) == "[a-df-i]"
    assert str(Charclass("^")) == "^"
    assert str(Charclass("\\")) == "\\\\"
    assert str(Charclass("a^")) == "[\\^a]"
    assert str(Charclass("0123456789a")) == "[0-9a]"
    assert str(Charclass("\t\v\r A")) == "[\\t\\v\\r A]"
    assert str(Charclass("\n\f A")) == "[\\n\\f A]"
    assert str(Charclass("\t\n\v\f\r A")) == "[\\t-\\r A]"
    assert (
        str(
            Charclass(
                "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz|"
            )
        )
        == "[0-9A-Z_a-z|]"
    )
    assert str(NONWORDCHAR) == "\\W"
    assert str(NONDIGITCHAR) == "\\D"
    assert str(NONSPACECHAR) == "\\S"
    assert str(DOT) == "."
    assert str(~Charclass("")) == "."
    assert str(~Charclass("a")) == "[^a]"
    assert str(~Charclass("{")) == "[^{]"
    assert str(~Charclass("\t")) == "[^\\t]"
    assert str(~Charclass("^")) == "[^\\^]"


def test_charclass_negation() -> None:
    assert ~~Charclass("a") == Charclass("a")
    assert Charclass("a") == ~~Charclass("a")


def test_charclass_union() -> None:
    # [ab] ∪ [bc] = [abc]
    assert Charclass("ab") | Charclass("bc") == Charclass("abc")
    # [ab] ∪ [^bc] = [^c]
    assert Charclass("ab") | ~Charclass("bc") == ~Charclass("c")
    # [^ab] ∪ [bc] = [^a]
    assert ~Charclass("ab") | Charclass("bc") == ~Charclass("a")
    # [^ab] ∪ [^bc] = [^b]
    assert ~Charclass("ab") | ~Charclass("bc") == ~Charclass("b")


def test_charclass_intersection() -> None:
    # [ab] ∩ [bc] = [b]
    assert Charclass("ab") & Charclass("bc") == Charclass("b")
    # [ab] ∩ [^bc] = [a]
    assert Charclass("ab") & ~Charclass("bc") == Charclass("a")
    # [^ab] ∩ [bc] = [c]
    assert ~Charclass("ab") & Charclass("bc") == Charclass("c")
    # [^ab] ∩ [^bc] = [^abc]
    assert ~Charclass("ab") & ~Charclass("bc") == ~Charclass("abc")

    assert (Charclass("ab") & Charclass("bcd") & Charclass("abcde")) == Charclass("b")


def test_empty() -> None:
    assert NULLCHARCLASS.empty()
    assert not DOT.empty()


def test_repartition_elementary() -> None:
    assert repartition([Charclass("a")]) == {
        Charclass("a"): [Charclass("a")],
    }


def test_repartition_elementary_2() -> None:
    assert repartition([Charclass("a"), ~Charclass("a")]) == {
        Charclass("a"): [Charclass("a")],
        ~Charclass("a"): [~Charclass("a")],
    }


def test_repartition_basic() -> None:
    assert repartition([Charclass("a"), Charclass("abc")]) == {
        Charclass("a"): [
            Charclass("a"),
        ],
        Charclass("abc"): [
            Charclass("a"),
            Charclass("bc"),
        ],
    }


def test_repartition_negation() -> None:
    assert repartition([Charclass("ab"), Charclass("a"), ~Charclass("ab")]) == {
        Charclass("ab"): [
            Charclass("a"),
            Charclass("b"),
        ],
        Charclass("a"): [
            Charclass("a"),
        ],
        ~Charclass("ab"): [
            ~Charclass("ab"),
        ],
    }


def test_repartition_negation_2() -> None:
    assert repartition([Charclass("ab"), Charclass("abc"), ~Charclass("ab")]) == {
        Charclass("ab"): [
            Charclass("ab"),
        ],
        Charclass("abc"): [
            Charclass("ab"),
            Charclass("c"),
        ],
        ~Charclass("ab"): [
            ~Charclass("abc"),
            Charclass("c"),
        ],
    }
    assert repartition(
        [
            ~Charclass("a"),
            ~Charclass("ab"),
            ~Charclass("abc"),
        ]
    ) == {
        ~Charclass("a"): [
            ~Charclass("abc"),
            Charclass("b"),
            Charclass("c"),
        ],
        ~Charclass("ab"): [
            ~Charclass("abc"),
            Charclass("c"),
        ],
        ~Charclass("abc"): [
            ~Charclass("abc"),
        ],
    }


def test_repartition_advanced() -> None:
    assert repartition(
        [
            Charclass("a"),
            Charclass("bcdef"),
            ~Charclass("abcdef"),
            Charclass("abcd"),
            ~Charclass("abcd"),
        ]
    ) == {
        Charclass("a"): [Charclass("a")],
        Charclass("bcdef"): [
            Charclass("bcd"),
            Charclass("ef"),
        ],
        ~Charclass("abcdef"): [
            ~Charclass("abcdef"),
        ],
        Charclass("abcd"): [
            Charclass("a"),
            Charclass("bcd"),
        ],
        ~Charclass("abcd"): [
            ~Charclass("abcdef"),
            Charclass("ef"),
        ],
    }


def test_repartition_advanced_2() -> None:
    assert repartition([WORDCHAR, DIGIT, DOT, NONDIGITCHAR, NULLCHARCLASS]) == {
        WORDCHAR: [
            DIGIT,
            Charclass("ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz"),
        ],
        DIGIT: [DIGIT],
        DOT: [
            ~Charclass((("0", "z"),)),
            DIGIT,
            Charclass(((":", "@"), ("[", "^"), ("`", "`"))),
            Charclass("ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz"),
        ],
        NONDIGITCHAR: [
            ~Charclass((("0", "z"),)),
            Charclass(((":", "@"), ("[", "^"), ("`", "`"))),
            Charclass("ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz"),
        ],
        NULLCHARCLASS: [
            # Yup, there's nothing here!
            # This should be impossible or at least cause no problems in practice
        ],
    }


# This should take a reasonable amount of time
# It was previously taking forever
def test_charclass_by_category() -> None:
    out = {}
    for i in range(0x101000):
        c = chr(i)
        cat = unicodedata.category(c)
        if cat not in out:
            out[cat] = [c]
        else:
            out[cat].append(c)
    for cat, cs in out.items():
        Charclass("".join(cs))



from __future__ import annotations

import pickle
import time

import pytest

from .charclass import Charclass
from .fsm import Fsm
from .parse import parse
from .rxelems import from_fsm

# pylint: disable=compare-to-empty-string
# pylint: disable=invalid-name
# pylint: disable=too-many-lines

if __name__ == "__main__":
    raise RuntimeError(
        "Test files can't be run directly. Use `python -m pytest greenery`"
    )


###############################################################################
# Stringification tests


def test_charclass_str() -> None:
    # Arbitrary ranges
    assert str(parse("[\\w:;<=>?@\\[\\\\\\]\\^`]")) == "[0-z]"
    # pylint: disable-next=fixme
    # TODO: what if \d is a proper subset of `chars`?

    # escape sequences are not preserved
    assert str(parse("\\x09")) == "\\t"

    # Printing ASCII control characters? You should get hex escapes
    assert str(parse("\\x00")) == "\\x00"


def test_parse_str_round_trip() -> None:
    # not "a[ab]b"
    assert str(parse("a.b")) == "a.b"
    assert str(parse("\\d{4}")) == "\\d{4}"
    assert str(parse("a.b()()")) == "a.b()()"


###############################################################################
# Test to_fsm() and alphabet-related functionality


def test_pattern_fsm() -> None:
    # "a[^a]"
    anota = parse("a[^a]").to_fsm()
    assert len(anota.states) == 4
    assert not anota.accepts("a")
    assert not anota.accepts("b")
    assert not anota.accepts("aa")
    assert anota.accepts("ab")
    assert anota.accepts("ac")
    assert not anota.accepts("ba")
    assert not anota.accepts("bb")

    # "0\\d"
    zeroD = parse("0\\d").to_fsm()
    assert zeroD.accepts("01")
    assert not zeroD.accepts("10")

    # "\\d{2}"
    d2 = parse("\\d{2}").to_fsm()
    assert not d2.accepts("")
    assert not d2.accepts("1")
    assert d2.accepts("11")
    assert not d2.accepts("111")

    # abc|def(ghi|jkl)
    conventional = parse("abc|def(ghi|jkl)").to_fsm()
    assert not conventional.accepts("a")
    assert not conventional.accepts("ab")
    assert conventional.accepts("abc")
    assert not conventional.accepts("abcj")
    assert conventional.accepts("defghi")
    assert conventional.accepts("defjkl")


def test_fsm() -> None:
    # You should be able to to_fsm() a single regular expression element
    # without supplying a specific alphabet. That should be determinable from
    # context.
    assert parse("a.b").to_fsm().accepts("acb")

    bad = parse("0{2}|1{2}").to_fsm()
    assert bad.accepts("00")
    assert bad.accepts("11")
    assert not bad.accepts("01")

    bad = parse("0{2}|1{2}").to_fsm()
    assert bad.accepts("00")
    assert bad.accepts("11")
    assert not bad.accepts("01")


def test_bug_28() -> None:
    # Starification was broken in FSMs
    assert not parse("(ab*)").to_fsm().star().accepts("bb")
    assert not parse("(ab*)*").to_fsm().accepts("bb")


###############################################################################
# Test matches(). Quite sparse at the moment


def test_wildcards_in_charclasses() -> None:
    # Allow "\w", "\d" and "\s" in `Charclass`es
    assert parse("[\\w~]*").matches("a0~")
    assert parse("[\\da]*").matches("0129a")
    assert parse("[\\s]+").matches(" \t \t ")


def test_block_comment_regex() -> None:
    # I went through several incorrect regexes for C block comments. Here we
    # show why the first few attempts were incorrect
    a = parse("/\\*(([^*]|\\*+[^*/])*)\\*/")
    assert a.matches("/**/")
    assert not a.matches("/***/")
    assert not a.matches("/****/")

    b = parse("/\\*(([^*]|\\*[^/])*)\\*/")
    assert b.matches("/**/")
    assert not b.matches("/***/")
    assert b.matches("/****/")

    c = parse("/\\*(([^*]|\\*+[^*/])*)\\*+/")
    assert c.matches("/**/")
    assert c.matches("/***/")
    assert c.matches("/****/")


def test_named_groups() -> None:
    a = parse("(?P<ng1>abc)")
    assert a.matches("abc")


def test_in() -> None:
    assert "a" in parse("a")
    assert "abcdsasda" in parse("\\w{4,10}")
    assert "abc" in parse("abc|def(ghi|jkl)")


###############################################################################
# Test string generators


def test_charclass_gen() -> None:
    gen = parse("[xyz]").strings()
    assert next(gen) == "x"
    assert next(gen) == "y"
    assert next(gen) == "z"

    with pytest.raises(StopIteration):
        next(gen)


def test_mult_gen() -> None:
    # One term
    gen = parse("[ab]").strings()
    assert next(gen) == "a"
    assert next(gen) == "b"

    with pytest.raises(StopIteration):
        next(gen)

    # No terms
    gen = parse("[ab]{0}").strings()
    assert next(gen) == ""

    with pytest.raises(StopIteration):
        next(gen)

    # Many terms
    gen = parse("[ab]*").strings()
    assert next(gen) == ""
    assert next(gen) == "a"
    assert next(gen) == "b"
    assert next(gen) == "aa"
    assert next(gen) == "ab"
    assert next(gen) == "ba"
    assert next(gen) == "bb"
    assert next(gen) == "aaa"


def test_conc_generator() -> None:
    gen = parse("[ab][cd]").strings()
    assert next(gen) == "ac"
    assert next(gen) == "ad"
    assert next(gen) == "bc"
    assert next(gen) == "bd"

    with pytest.raises(StopIteration):
        next(gen)


def test_pattern_generator() -> None:
    gen = parse("[ab]|[cde]").strings()
    assert next(gen) == "a"
    assert next(gen) == "b"
    assert next(gen) == "c"
    assert next(gen) == "d"
    assert next(gen) == "e"

    with pytest.raises(StopIteration):
        next(gen)

    # more complex
    gen = parse("abc|def(ghi|jkl)").strings()
    assert next(gen) == "abc"
    assert next(gen) == "defghi"
    assert next(gen) == "defjkl"

    gen = parse("[0-9a-fA-F]{3,10}").strings()
    assert next(gen) == "000"
    assert next(gen) == "001"
    assert next(gen) == "002"


def test_infinite_generation() -> None:
    # Infinite generator, flummoxes both depth-first and breadth-first searches
    gen = parse("a*b*").strings()
    assert next(gen) == ""
    assert next(gen) == "a"
    assert next(gen) == "b"
    assert next(gen) == "aa"
    assert next(gen) == "ab"
    assert next(gen) == "bb"
    assert next(gen) == "aaa"
    assert next(gen) == "aab"
    assert next(gen) == "abb"
    assert next(gen) == "bbb"
    assert next(gen) == "aaaa"


def test_wildcard_generator() -> None:
    # Generator needs to handle wildcards as well. Wildcards come last.
    gen = parse("a.b").strings(otherchar="*")
    assert next(gen) == "aab"
    assert next(gen) == "abb"
    assert next(gen) == "a*b"

    with pytest.raises(StopIteration):
        next(gen)


def test_forin() -> None:
    assert tuple(parse("abc|def(ghi|jkl)")) == ("abc", "defghi", "defjkl")


###############################################################################
# Test cardinality() and len()


def test_cardinality() -> None:
    # pylint: disable-next=compare-to-zero
    assert parse("[]").cardinality() == 0
    assert parse("[ab]{3}").cardinality() == 8
    assert parse("[ab]{2,3}").cardinality() == 12
    assert len(parse("abc|def(ghi|jkl)")) == 3
    with pytest.raises(OverflowError):
        len(parse(".*"))


def test_cardinality_harder() -> None:
    assert parse("[]?").cardinality() == 1
    assert parse("[]{0,6}").cardinality() == 1
    assert parse("[ab]{0,3}").cardinality() == 15


###############################################################################


def test_copy() -> None:
    x = parse("abc|def(ghi|jkl)")
    assert x.copy() == x


###############################################################################
# Test from_fsm()


def test_dot() -> None:
    # not "a[ab]b"
    assert str(from_fsm(parse("a.b").to_fsm())) == "a.b"


def test_abstar() -> None:
    # Buggggs.
    abstar = Fsm(
        alphabet={Charclass("a"), ~Charclass("ab"), Charclass("b")},
        states={0, 1},
        initial=0,
        finals={0},
        map={
            0: {Charclass("a"): 0, ~Charclass("ab"): 1, Charclass("b"): 0},
            1: {Charclass("a"): 1, ~Charclass("ab"): 1, Charclass("b"): 1},
        },
    )
    assert str(from_fsm(abstar)) == "[ab]*"


def test_adotb() -> None:
    adotb = Fsm(
        alphabet={Charclass("a"), ~Charclass("ab"), Charclass("b")},
        states={0, 1, 2, 3, 4},
        initial=0,
        finals={4},
        map={
            0: {Charclass("a"): 2, ~Charclass("ab"): 1, Charclass("b"): 1},
            1: {Charclass("a"): 1, ~Charclass("ab"): 1, Charclass("b"): 1},
            2: {Charclass("a"): 3, ~Charclass("ab"): 3, Charclass("b"): 3},
            3: {Charclass("a"): 1, ~Charclass("ab"): 1, Charclass("b"): 4},
            4: {Charclass("a"): 1, ~Charclass("ab"): 1, Charclass("b"): 1},
        },
    )
    assert str(from_fsm(adotb)) == "a.b"


def test_rxelems_recursion_error() -> None:
    # Catch a recursion error
    assert (
        str(
            from_fsm(
                Fsm(
                    alphabet={Charclass("0"), Charclass("1"), ~Charclass("01")},
                    states={0, 1, 2, 3},
                    initial=3,
                    finals={1},
                    map={
                        0: {Charclass("0"): 1, Charclass("1"): 1, ~Charclass("01"): 2},
                        1: {Charclass("0"): 2, Charclass("1"): 2, ~Charclass("01"): 2},
                        2: {Charclass("0"): 2, Charclass("1"): 2, ~Charclass("01"): 2},
                        3: {Charclass("0"): 0, Charclass("1"): 2, ~Charclass("01"): 2},
                    },
                )
            )
        )
        == "0[01]"
    )


def test_even_star_bug1() -> None:
    # Bug fix. This is a(a{2})* (i.e. accepts an odd number of "a" chars in a
    # row), but when from_fsm() is called, the result is "a+". Turned out to be
    # a fault in the rxelems.multiplier.__mul__() routine
    elesscomplex = Fsm(
        alphabet={Charclass("a"), ~Charclass("a")},
        states={0, 1, 2},
        initial=0,
        finals={1},
        map={
            0: {Charclass("a"): 1, ~Charclass("a"): 2},
            1: {Charclass("a"): 0, ~Charclass("a"): 2},
            2: {Charclass("a"): 2, ~Charclass("a"): 2},
        },
    )
    assert not elesscomplex.accepts("")
    assert elesscomplex.accepts("a")
    assert not elesscomplex.accepts("aa")
    assert elesscomplex.accepts("aaa")
    elesscomplex_pat = from_fsm(elesscomplex)
    assert str(elesscomplex_pat) in {"a(a{2})*", "(a{2})*a"}
    elesscomplex = elesscomplex_pat.to_fsm()
    assert not elesscomplex.accepts("")
    assert elesscomplex.accepts("a")
    assert not elesscomplex.accepts("aa")
    assert elesscomplex.accepts("aaa")
    gen = elesscomplex.strings([])
    assert next(gen) == "a"
    assert next(gen) == "aaa"
    assert next(gen) == "aaaaa"
    assert next(gen) == "aaaaaaa"


def test_binary_3() -> None:
    # Binary numbers divisible by 3.
    # Disallows the empty string
    # Allows "0" on its own, but not leading zeroes.
    div3 = from_fsm(
        Fsm(
            alphabet={Charclass("0"), Charclass("1"), ~Charclass("01")},
            states={-2, -1, 0, 1, 2, 3},
            initial=-2,
            finals={-1, 0},
            map={
                -2: {Charclass("0"): -1, Charclass("1"): 1, ~Charclass("01"): 3},
                -1: {Charclass("0"): 3, Charclass("1"): 3, ~Charclass("01"): 3},
                0: {Charclass("0"): 0, Charclass("1"): 1, ~Charclass("01"): 3},
                1: {Charclass("0"): 2, Charclass("1"): 0, ~Charclass("01"): 3},
                2: {Charclass("0"): 1, Charclass("1"): 2, ~Charclass("01"): 3},
                3: {Charclass("0"): 3, Charclass("1"): 3, ~Charclass("01"): 3},
            },
        )
    )
    assert str(parse("(0|1)").reduce()) == "[01]"
    assert str(parse("(0|12)").reduce()) == "0|12"
    assert str(parse("(0|1(01*0|10*1)*10*)").reduce()) == "0|1(01*0|10*1)*10*"
    assert str(div3) == "0|1(01*0|10*1)*10*"
    gen = div3.strings()
    assert next(gen) == "0"
    assert next(gen) == "11"
    assert next(gen) == "110"
    assert next(gen) == "1001"
    assert next(gen) == "1100"


def test_base_N() -> None:
    # Machine accepts only numbers in selected base (e.g. 2, 10) that are
    # divisible by N (e.g. 3, 7).
    # "0" alone is acceptable, but leading zeroes (e.g. "00", "07") are not
    base = 2
    N = 3
    assert base <= 10
    anything_else = ~Charclass("".join(str(i) for i in range(base)))
    divN = from_fsm(
        Fsm(
            alphabet=({Charclass(str(i)) for i in range(base)} | {anything_else}),
            states=frozenset(range(N)) | {-2, -1, N},
            initial=-2,
            finals={-1, 0},
            map={
                -2: {
                    Charclass("0"): -1,
                    **{Charclass(str(j)): j % N for j in range(1, base)},
                    anything_else: N,
                },
                -1: {
                    **{Charclass(str(j)): N for j in range(base)},
                    anything_else: N,
                },
                **{
                    i: {
                        **{Charclass(str(j)): (i * base + j) % N for j in range(base)},
                        anything_else: N,
                    }
                    for i in range(N)
                },
                N: {
                    **{Charclass(str(j)): N for j in range(base)},
                    anything_else: N,
                },
            },
        )
    )
    gen = divN.strings()
    a = next(gen)
    assert a == "0"
    for i in range(7):
        b = next(gen)
        assert int(a, base) + N == int(b, base)
        a = b


def test_dead_default() -> None:
    blockquote = from_fsm(
        Fsm(
            alphabet={Charclass("/"), Charclass("*"), ~Charclass("/*")},
            states={0, 1, 2, 3, 4, 5},
            initial=0,
            finals={4},
            map={
                0: {Charclass("/"): 1, ~Charclass("/*"): 5, Charclass("*"): 5},
                1: {Charclass("/"): 5, ~Charclass("/*"): 5, Charclass("*"): 2},
                2: {Charclass("/"): 2, ~Charclass("/*"): 2, Charclass("*"): 3},
                3: {Charclass("/"): 4, ~Charclass("/*"): 2, Charclass("*"): 3},
                4: {Charclass("/"): 5, ~Charclass("/*"): 5, Charclass("*"): 5},
                5: {Charclass("/"): 5, ~Charclass("/*"): 5, Charclass("*"): 5},
            },
        )
    )
    assert str(blockquote) == "/\\*([^*]|\\*+[^*/])*\\*+/"


###############################################################################
# charclass set operations


def test_charclass_union() -> None:
    assert (parse("[ab]") | parse("[bc]")).reduce() == parse("[abc]")
    assert (parse("[ab]") | parse("[^bc]")).reduce() == parse("[^c]")
    assert (parse("[^ab]") | parse("[bc]")).reduce() == parse("[^a]")
    assert (parse("[^ab]") | parse("[^bc]")).reduce() == parse("[^b]")


def test_charclass_intersection() -> None:
    assert parse("[ab]") & parse("[bc]") == parse("b")
    assert parse("[ab]") & parse("[^bc]") == parse("a")
    assert parse("[^ab]") & parse("[bc]") == parse("c")
    assert parse("[^ab]") & parse("[^bc]") == parse("[^abc]")


###############################################################################
# Emptiness detection


def test_empty() -> None:
    assert not parse("a{0}").empty()
    assert parse("[]").empty()
    assert not parse("[]?").empty()
    assert parse("a[]").empty()
    assert not parse("a[]?").empty()
    assert not parse("a{0}").empty()
    assert not parse("[]?").empty()


###############################################################################
# Test everythingbut()


def test_everythingbut() -> None:
    # Regexes are usually gibberish but we make a few claims
    a = parse("a")
    notA = a.everythingbut().to_fsm()
    assert notA.accepts("")
    assert not notA.accepts("a")
    assert notA.accepts("aa")

    # everythingbut(), called twice, should take us back to where we started.
    beer = parse("beer")
    notBeer = beer.everythingbut()
    beer2 = notBeer.everythingbut()
    assert str(beer2) == "be{2}r"

    # ".*" becomes "[]" and vice versa under this call.
    assert str(parse(".*").everythingbut()) == "[]"
    assert str(parse("[]").everythingbut()) == ".*"


def test_isinstance_bug() -> None:
    # Problem relating to isinstance(). The class `Mult` was occurring as both
    # rxelems.Mult and as __main__.Mult and apparently these count as different
    # classes for some reason, so isinstance(m, Mult) was returning false.
    var = str(parse("").everythingbut()) + "aa" + str(parse("").everythingbut())
    assert var == ".+aa.+"

    starfree = parse(var).everythingbut()
    assert str(starfree) == "(.(a?[^a])*a{0,2})?"


###############################################################################


def test_equivalence() -> None:
    assert parse("aa*").equivalent(parse("a*a"))
    assert parse("([ab]*a|[bc]*c)?b*").equivalent(parse("b*(a[ab]*|c[bc]*)?"))


###############################################################################
# Test reversed()


def test_regex_reversal() -> None:
    assert parse("b").reversed() == parse("b")
    assert parse("e*").reversed() == parse("e*")
    assert parse("bear").reversed() == parse("raeb")
    assert parse("beer").reversed() == parse("reeb")
    assert parse("abc|def|ghi").reversed() == parse("cba|fed|ihg")
    assert parse("(abc)*d").reversed() == parse("d(cba)*")


###############################################################################
# Tests for some more set operations


def test_set_ops() -> None:
    assert parse("[abcd]") - parse("a") == parse("[bcd]")
    assert parse("[abcd]") ^ parse("[cdef]") == parse("[abef]")


###############################################################################
# Test methods for finding common parts of regular expressions.


def test_pattern_commonconc() -> None:
    # pylint: disable=protected-access
    assert str(parse("aa|aa")._commonconc()) == "aa"
    assert str(parse("abc|aa")._commonconc()) == "a"
    assert str(parse("a|bc")._commonconc()) == ""
    assert str(parse("cf{1,2}|cf")._commonconc()) == "cf"
    assert str(parse("ZA|ZB|ZC")._commonconc()) == "Z"
    assert str(parse("Z+A|ZB|ZZC")._commonconc()) == "Z"
    assert str(parse("a{2}b|a+c")._commonconc()) == "a"


def test_pattern_commonconc_suffix() -> None:
    # pylint: disable=protected-access
    assert str(parse("a|bc")._commonconc(suffix=True)) == ""
    assert str(parse("aa|bca")._commonconc(suffix=True)) == "a"
    assert str(parse("xyza|abca|a")._commonconc(suffix=True)) == "a"
    assert str(parse("f{2,3}c|fc")._commonconc(suffix=True)) == "fc"
    assert str(parse("aa")._commonconc(suffix=True)) == "aa"


###############################################################################
# Basic concatenation reduction tests


def test_reduce_concatenations() -> None:
    assert str(parse("aa").reduce()) == "a{2}"
    assert str(parse("bb").reduce()) == "b{2}"
    assert str(parse("b*b").reduce()) == "b+"
    assert str(parse("aa{2,}").reduce()) == "a{3,}"
    assert str(parse("a*a{2}").reduce()) == "a{2,}"
    assert str(parse("aa{0,8}").reduce()) == "a{1,9}"
    assert str(parse("b{0,8}b").reduce()) == "b{1,9}"
    assert str(parse("aab").reduce()) == "a{2}b"
    assert str(parse("abb").reduce()) == "ab{2}"
    assert str(parse("abb*").reduce()) == "ab+"
    assert str(parse("abbc").reduce()) == "ab{2}c"
    assert str(parse("a?ab").reduce()) == "a{1,2}b"
    assert str(parse("(ac{2}|bc+)c").reduce()) == "(ac|bc*)c{2}"
    assert str(parse("a(a{2}b|a+c)").reduce()) == "a{2}(a*c|ab)"
    assert str(parse("a{2,3}(a{2}b|a+c)").reduce()) == "a{3,4}(a*c|ab)"
    assert str(parse("(ba{2}|ca+)a{2,3}").reduce()) == "(ba|ca*)a{3,4}"
    assert str(parse("za{2,3}(a{2}b|a+c)").reduce()) == "za{3,4}(a*c|ab)"
    assert str(parse("(ba{2}|ca+)a{2,3}z").reduce()) == "(ba|ca*)a{3,4}z"
    assert str(parse("(a|bc)(a|bc)").reduce()) == "(a|bc){2}"
    assert str(parse("a+[ab]+").reduce()) == "a[ab]+"
    assert str(parse("a{3,8}[ab]+").reduce()) == "a{3}[ab]+"
    assert str(parse("[ab]+b+").reduce()) == "[ab]+b"
    assert str(parse("[ab]+a{3,8}").reduce()) == "[ab]+a{3}"
    assert str(parse("\\d+\\w+").reduce()) == "\\d\\w+"
    assert str(parse("[ab]+a?").reduce()) == "[ab]+"


###############################################################################
# Multiplication tests


def test_mult_multiplication() -> None:
    assert parse("(a{2,3}){1,1}").reduce() == parse("a{2,3}").reduce()
    assert parse("(a{2,3}){1}").reduce() == parse("a{2,3}").reduce()
    assert parse("(a{2,3})").reduce() == parse("a{2,3}").reduce()
    assert parse("(a{2,3}){4,5}").reduce() == parse("a{8,15}").reduce()
    assert parse("(a{2,}){2,}").reduce() == parse("a{4,}").reduce()


def test_even_star_bug2() -> None:
    # Defect: (a{2})* should NOT reduce to a*
    assert parse("(a{2})*").reduce() != parse("a*").reduce()


def test_two_or_more_qm_bug() -> None:
    assert str(parse("(a{2,})?").reduce()) == "(a{2,})?"


def test_two_two_bug() -> None:
    assert str(parse("(a{2}){2}").reduce()) == "a{4}"


###############################################################################
# Test intersection (&)


def test_mult_intersection() -> None:
    assert str(parse("a") & parse("a")) == "a"
    assert str(parse("a*") & parse("a")) == "a"
    assert str(parse("a") & parse("a?")) == "a"
    assert str(parse("a{2}") & parse("a{2,}")) == "a{2}"
    assert str(parse("a*") & parse("a+")) == "a+"
    assert str(parse("a{2}") & parse("a{4}")) == "[]"
    assert str(parse("a{3,}") & parse("a{3,}")) == "a{3,}"


def test_parse_regex_intersection() -> None:
    assert str(parse("a*") & parse("b*")) == ""
    assert str(parse("a") & parse("b")) == "[]"
    assert str(parse("\\d") & parse(".")) == "\\d"
    assert str(parse("\\d{2}") & parse("0.")) == "0\\d"
    assert str(parse("\\d{2}") & parse("19.*")) == "19"
    assert str(parse("\\d{3}") & parse("19.*")) == "19\\d"
    assert str(parse("abc...") & parse("...def")) == "abcdef"
    assert str(parse("[bc]*[ab]*") & parse("[ab]*[bc]*")) in {
        "([ab]*a|[bc]*c)?b*",
        "b*(a[ab]*|c[bc]*)?",
    }
    assert str(parse("\\W*")) == "\\W*"
    assert str(parse("[a-g0-8$%\\^]+")) == "[$%0-8\\^a-g]+"
    assert str(parse("[^d]{2,8}")) == "[^d]{2,8}"
    assert str(parse("\\W*") & parse("[a-g0-8$%\\^]+")) == "[$%\\^]+"
    assert str(parse("[ab]{1,2}") & parse("[^a]{1,2}")) == "b{1,2}"
    assert str(parse("[ab]?") & parse("[^a]?")) == "b?"
    assert parse("a{0,2}").matches("")
    assert parse("[ab]{0,2}").matches("")
    assert parse("[^a]{0,2}").matches("")
    assert parse("b{0,2}").matches("")
    assert str(parse("[ab]{0,2}") & parse("[^a]{0,2}")) == "b{0,2}"
    assert str(parse("[ab]{0,4}") & parse("[^a]{0,4}")) == "b{0,4}"
    assert str(parse("[abc]{0,8}") & parse("[^a]{0,8}")) == "[bc]{0,8}"
    assert (
        str(parse("[a-g0-8$%\\^]{0,8}") & parse("[^d]{0,8}")) == "[$%0-8\\^abcefg]{0,8}"
    )
    assert str(parse("[a-g0-8$%\\^]+") & parse("[^d]{0,8}")) == "[$%0-8\\^abcefg]{1,8}"
    assert str(parse("[a-g0-8$%\\^]+") & parse("[^d]{2,8}")) == "[$%0-8\\^abcefg]{2,8}"
    assert (
        str(parse("\\W*") & parse("[a-g0-8$%\\^]+") & parse("[^d]{2,8}"))
        == "[$%\\^]{2,8}"
    )
    assert (
        str(parse("\\d{4}-\\d{2}-\\d{2}") & parse("19.*")) == "19\\d{2}-\\d{2}-\\d{2}"
    )


def test_complexify() -> None:
    # Complexify!
    gen = (parse("[bc]*[ab]*") & parse("[ab]*[bc]*")).strings()
    assert next(gen) == ""
    assert next(gen) == "a"
    assert next(gen) == "b"
    assert next(gen) == "c"
    assert next(gen) == "aa"
    assert next(gen) == "ab"
    # no "ac"
    assert next(gen) == "ba"
    assert next(gen) == "bb"
    assert next(gen) == "bc"
    # no "ca"
    assert next(gen) == "cb"
    assert next(gen) == "cc"
    assert next(gen) == "aaa"


def test_silly_reduction() -> None:
    # This one is horrendous and we have to jump through some hoops to get to
    # a sensible result. Probably not a good unit test actually.
    long = (
        "(aa|bb*aa)a*|((ab|bb*ab)|(aa|bb*aa)a*b)"
        + "((ab|bb*ab)|(aa|bb*aa)a*b)*"
        + "(aa|bb*aa)a*|((ab|bb*ab)|(aa|bb*aa)a*b)"
        + "((ab|bb*ab)|(aa|bb*aa)a*b)*"
    )
    long_pat1 = parse(long)
    long_fsm = long_pat1.to_fsm().reversed()
    long_pat2 = from_fsm(long_fsm).reversed()
    assert str(long_pat2) == "[ab]*a[ab]"
    short = "[ab]*a?b*|[ab]*b?a*"
    assert str(parse(".*") & parse(short)) == "[ab]*"


###############################################################################
# reduce() tests


def test_mult_reduction_easy() -> None:
    assert str(parse("a").reduce()) == "a"
    assert str(parse("a").reduce()) == "a"
    assert str(parse("a?").reduce()) == "a?"
    assert str(parse("a{0}").reduce()) == ""
    assert str(parse("[]").reduce()) == "[]"
    assert str(parse("[]?").reduce()) == ""
    assert str(parse("[]{0}").reduce()) == ""
    assert str(parse("[]{0,5}").reduce()) == ""


def test_conc_reduction_basic() -> None:
    assert str(parse("a").reduce()) == "a"
    assert str(parse("a{3,4}").reduce()) == "a{3,4}"
    assert str(parse("ab").reduce()) == "ab"
    assert str(parse("a[]b").reduce()) == "[]"


def test_pattern_reduce_basic() -> None:
    assert str(parse("ab|cd").reduce()) == "ab|cd"
    assert str(parse("a{2}b{2}").reduce()) == "a{2}b{2}"
    assert str(parse("a{2}").reduce()) == "a{2}"
    assert str(parse("a").reduce()) == "a"
    assert str(parse("(((a)))").reduce()) == "a"


def test_empty_conc_suppression() -> None:
    assert str(parse("[]0\\d").reduce()) == "[]"


def test_nested_pattern_reduction() -> None:
    # a(d(ab|a*c)) -> ad(ab|a*c)
    assert str(parse("a(d(ab|a*c))").reduce()) == "ad(a*c|ab)"


def test_mult_factor_out_qm() -> None:
    # `Mult` contains a `Pattern` containing an empty `Conc`? Pull the empty
    # part out where it's external
    assert str(parse("a|b*|").reduce()) == "a|b*"
    assert str(parse("(a|b*|)").reduce()) == "a|b*"
    assert str(parse("(a|b*|)c").reduce()) == "(a|b*)c"
    # This happens even if `EMPTYSTRING` is the only thing left inside the
    # `Mult`
    assert str(parse("()").reduce()) == ""
    assert str(parse("([$%\\^]|){1}").reduce()) == "[$%\\^]?"


def test_remove_unnecessary_parens() -> None:
    # `Mult` contains a `Pattern` containing a single `Conc` containing a
    # single `Mult`? That can be reduced greatly
    assert str(parse("(a){2}b").reduce()) == "a{2}b"
    assert str(parse("(a?)+b").reduce()) == "a*b"
    assert str(parse("([ab])*").reduce()) == "[ab]*"
    assert str(parse("(c{1,2}){3,4}").reduce()) == "c{3,8}"


def test_obvious_reduction() -> None:
    assert str(parse("(a|b)*").reduce()) == "[ab]*"


def test_mult_squoosh() -> None:
    # sequence squooshing of mults within a `Conc`
    assert str(parse("[0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]").reduce()) == "[0-9A-Fa-f]{3}"
    assert str(parse("[$%\\^]?[$%\\^]").reduce()) == "[$%\\^]{1,2}"
    assert (
        str(
            parse(
                "(|(|(|(|(|[$%\\^])[$%\\^])[$%\\^])[$%\\^])[$%\\^])[$%\\^][$%\\^]"
            ).reduce()
        )
        == "[$%\\^]{2,7}"
    )


def test_bad_reduction_bug() -> None:
    # DEFECT: "0{2}|1{2}" was erroneously reduced() to "[01]{2}"
    assert parse("0{2}|1{2}").reduce() != parse("[01]{2}")
    assert parse("0|[1-9]|ab").reduce() == parse("\\d|ab")
    assert parse("0|[1-9]|a{5,7}").reduce() == parse("\\d|a{5,7}")
    assert parse("0|(0|[1-9]|a{5,7})").reduce() == parse("0|(\\d|a{5,7})")
    # pylint: disable-next=fixme
    # TODO: should do better than this! Merge that 0


def test_common_prefix_pattern_reduction() -> None:
    assert str(parse("a{2}b|a+c").reduce()) == "a(a*c|ab)"


def test_epsilon_reduction() -> None:
    assert str(parse("|(ab)*|def").reduce()) == "(ab)*|def"
    assert str(parse("|(ab)+|def").reduce()) == "(ab)*|def"
    assert str(parse("|.+").reduce()) == ".*"
    assert str(parse("|a+|b+").reduce()) in {"a+|b*", "a*|b+"}


def test_charclass_intersection_2() -> None:
    assert (parse("[A-z]") & parse("[^g]")).reduce() == parse("[A-fh-z]").reduce()


def test_reduce_boom() -> None:
    # make sure recursion problem in reduce() has gone away
    assert str(parse("([1-9]|0)").reduce()) == "\\d"


def test_new_reduce() -> None:
    # The @reduce_after decorator has been removed from many methods since it
    # takes unnecessary time which the user may not wish to spend.
    # This alters the behaviour of several methods and also exposes a new
    # opportunity for `Conc.reduce()`
    assert str(parse("a()").reduce()) == "a"
    assert str(parse("a()()").reduce()) == "a"
    assert str(parse("a.b()()").reduce()) == "a.b"


def test_main_bug() -> None:
    assert str(parse("a*").reduce()) == "a*"
    assert str(parse("a|a*").reduce()) == "a*"
    assert str(parse("a{1,2}|a{3,4}|bc").reduce()) == "a{1,4}|bc"
    assert str(parse("a{1,2}|bc|a{3,4}").reduce()) == "a{1,4}|bc"
    assert str(parse("a{1,2}|a{3,4}|a{5,6}|bc").reduce()) == "a{1,6}|bc"
    assert str(parse("a{1,2}|a{3}|a{5,6}").reduce()) == "a{1,2}(a?|a{4})"
    assert str(parse("a{1,2}|a{3}|a{5,6}|bc").reduce()) == "a{1,3}|a{5,6}|bc"
    assert str(parse("a{1,2}|a{4}|a{5,6}").reduce()) == "a{1,2}(a{3,4})?"
    assert str(parse("a{1,2}|a{4}|a{5,6}|bc").reduce()) == "a{1,2}|a{4,6}|bc"
    assert str((parse("a") | parse("a*")).reduce()) == "a*"


def test_bug_28_b() -> None:
    # Defect in rxelems.to_fsm()
    assert not parse("(ab*)*").to_fsm().accepts("bb")


def test_derive() -> None:
    assert str(parse("a+").derive("a")) == "a*"
    assert str(parse("a+|b+").derive("a")) == "a*"
    assert str(parse("abc|ade").derive("a")) == "bc|de"
    assert str(parse("abc|ade").derive("ab")) == "c"
    assert str(parse("abc|ade").derive("c")) == "[]"


def test_bug_36_1() -> None:
    etc1 = parse(".*").to_fsm()
    etc2 = parse("s.*").to_fsm()
    assert etc1.accepts("s")
    assert etc2.accepts("s")
    assert not etc1.isdisjoint(etc2)
    assert not etc2.isdisjoint(etc1)


def test_bug_36_2() -> None:
    etc1 = parse("/etc/.*").to_fsm()
    etc2 = parse("/etc/something.*").to_fsm()
    assert etc1.accepts("/etc/something")
    assert etc2.accepts("/etc/something")
    assert not etc1.isdisjoint(etc2)
    assert not etc2.isdisjoint(etc1)


def test_isdisjoint() -> None:
    xyzzy = parse("xyz(zy)?")
    xyz = parse("xyz")
    blippy = parse("blippy")
    assert xyzzy.isdisjoint(blippy)
    assert not xyzzy.isdisjoint(xyz)


def test_bug_slow() -> None:
    # issue #43
    m = Fsm(
        alphabet={
            Charclass("R"),
            Charclass("L"),
            Charclass("U"),
            Charclass("D"),
            ~Charclass("RLUD"),
        },
        states={
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
            20,
            21,
        },
        initial=0,
        finals={20},
        map={
            0: {
                Charclass("R"): 21,
                Charclass("L"): 21,
                Charclass("U"): 2,
                Charclass("D"): 1,
                ~Charclass("RLUD"): 21,
            },
            1: {
                Charclass("R"): 21,
                Charclass("L"): 3,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            2: {
                Charclass("R"): 21,
                Charclass("L"): 4,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            3: {
                Charclass("R"): 21,
                Charclass("L"): 21,
                Charclass("U"): 5,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            4: {
                Charclass("R"): 21,
                Charclass("L"): 21,
                Charclass("U"): 21,
                Charclass("D"): 6,
                ~Charclass("RLUD"): 21,
            },
            5: {
                Charclass("R"): 7,
                Charclass("L"): 21,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            6: {
                Charclass("R"): 8,
                Charclass("L"): 21,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            7: {
                Charclass("R"): 21,
                Charclass("L"): 21,
                Charclass("U"): 9,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            8: {
                Charclass("R"): 21,
                Charclass("L"): 21,
                Charclass("U"): 21,
                Charclass("D"): 10,
                ~Charclass("RLUD"): 21,
            },
            9: {
                Charclass("R"): 21,
                Charclass("L"): 11,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            10: {
                Charclass("R"): 21,
                Charclass("L"): 12,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            11: {
                Charclass("R"): 21,
                Charclass("L"): 13,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            12: {
                Charclass("R"): 21,
                Charclass("L"): 14,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            13: {
                Charclass("R"): 21,
                Charclass("L"): 21,
                Charclass("U"): 21,
                Charclass("D"): 15,
                ~Charclass("RLUD"): 21,
            },
            14: {
                Charclass("R"): 21,
                Charclass("L"): 21,
                Charclass("U"): 16,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            15: {
                Charclass("R"): 17,
                Charclass("L"): 21,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            16: {
                Charclass("R"): 18,
                Charclass("L"): 21,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            17: {
                Charclass("R"): 21,
                Charclass("L"): 21,
                Charclass("U"): 21,
                Charclass("D"): 19,
                ~Charclass("RLUD"): 21,
            },
            18: {
                Charclass("R"): 21,
                Charclass("L"): 21,
                Charclass("U"): 19,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            19: {
                Charclass("R"): 21,
                Charclass("L"): 20,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            20: {
                Charclass("R"): 21,
                Charclass("L"): 21,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
            21: {
                Charclass("R"): 21,
                Charclass("L"): 21,
                Charclass("U"): 21,
                Charclass("D"): 21,
                ~Charclass("RLUD"): 21,
            },
        },
    )
    t1 = time.time()
    ll = from_fsm(m)
    t2 = time.time()
    assert (t2 - t1) < 60  # should finish in way under 1s
    assert ll == parse("(DLURULLDRD|ULDRDLLURU)L").reduce()


def test_bug_48_simpler() -> None:
    assert (
        str(
            from_fsm(
                Fsm(
                    alphabet={Charclass("d"), ~Charclass("d")},
                    states={0, 1, 2},
                    initial=0,
                    finals={1},
                    map={
                        0: {Charclass("d"): 1, ~Charclass("d"): 2},
                        1: {Charclass("d"): 2, ~Charclass("d"): 2},
                        2: {Charclass("d"): 2, ~Charclass("d"): 2},
                    },
                )
            )
        )
        == "d"
    )


def test_bug_48() -> None:
    machine = Fsm(
        alphabet={
            Charclass("_"),
            Charclass("a"),
            Charclass("d"),
            Charclass("e"),
            Charclass("g"),
            Charclass("m"),
            Charclass("n"),
            Charclass("o"),
            Charclass("p"),
            ~Charclass("_adegmnop"),
        },
        states={0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13},
        initial=0,
        finals={12},
        map={
            0: {
                Charclass("_"): 13,
                Charclass("a"): 13,
                Charclass("d"): 1,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 13,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            1: {
                Charclass("_"): 13,
                Charclass("a"): 2,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 13,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            2: {
                Charclass("_"): 13,
                Charclass("a"): 13,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 3,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            3: {
                Charclass("_"): 13,
                Charclass("a"): 4,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 13,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            4: {
                Charclass("_"): 13,
                Charclass("a"): 13,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 5,
                Charclass("m"): 13,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            5: {
                Charclass("_"): 13,
                Charclass("a"): 13,
                Charclass("d"): 13,
                Charclass("e"): 6,
                Charclass("g"): 13,
                Charclass("m"): 13,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            6: {
                Charclass("_"): 7,
                Charclass("a"): 13,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 13,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            7: {
                Charclass("_"): 13,
                Charclass("a"): 13,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 13,
                Charclass("n"): 13,
                Charclass("o"): 8,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            8: {
                Charclass("_"): 13,
                Charclass("a"): 13,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 13,
                Charclass("n"): 9,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            9: {
                Charclass("_"): 10,
                Charclass("a"): 13,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 13,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            10: {
                Charclass("_"): 13,
                Charclass("a"): 13,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 11,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            11: {
                Charclass("_"): 13,
                Charclass("a"): 13,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 13,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 12,
                ~Charclass("_adegmnop"): 13,
            },
            12: {
                Charclass("_"): 13,
                Charclass("a"): 13,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 13,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
            13: {
                Charclass("_"): 13,
                Charclass("a"): 13,
                Charclass("d"): 13,
                Charclass("e"): 13,
                Charclass("g"): 13,
                Charclass("m"): 13,
                Charclass("n"): 13,
                Charclass("o"): 13,
                Charclass("p"): 13,
                ~Charclass("_adegmnop"): 13,
            },
        },
    )

    rex = from_fsm(machine)
    assert str(rex) == "damage_on_mp"


def test_pickle() -> None:
    f1 = parse("a{0,4}").to_fsm()
    f2 = parse("a{0,3}").to_fsm()

    assert f2 < f1

    f1_unpickled = pickle.loads(pickle.dumps(f1))
    f2_unpickled = pickle.loads(pickle.dumps(f2))

    assert f2_unpickled < f1_unpickled



from __future__ import annotations

import pytest

from .bound import Bound
from .charclass import Charclass
from .multiplier import ONE, PLUS, QM, STAR, ZERO, Multiplier
from .rxelems import Conc, Mult


def test_conc_equality() -> None:
    a = Conc(Mult(Charclass("a"), ONE))
    assert a == Conc(Mult(Charclass("a"), ONE))
    assert a != Conc(Mult(Charclass("b"), ONE))
    assert a != Conc(Mult(Charclass("a"), QM))
    assert a != Conc(Mult(Charclass("a"), Multiplier(Bound(1), Bound(2))))
    assert a != Conc()


def test_conc_str() -> None:
    assert (
        str(
            Conc(
                Mult(Charclass("a"), ONE),
                Mult(Charclass("b"), ONE),
                Mult(Charclass("c"), ONE),
                Mult(Charclass("d"), ONE),
                Mult(Charclass("e"), ONE),
                Mult(~Charclass("fg"), STAR),
                Mult(Charclass("h"), Multiplier(Bound(5), Bound(5))),
                Mult(Charclass("abcdefghijklmnopqrstuvwxyz"), PLUS),
            )
        )
        == "abcde[^fg]*h{5}[a-z]+"
    )


def test_conc_common() -> None:
    a = Mult(Charclass("A"), ONE)
    b = Mult(Charclass("B"), ONE)
    c = Mult(Charclass("C"), ONE)
    y = Mult(Charclass("y"), ONE)
    z = Mult(Charclass("Z"), ONE)
    zstar = Mult(Charclass("Z"), STAR)

    assert Conc(a, a, z, y).common(Conc(b, b, z, y), suffix=True) == Conc(z, y)
    assert Conc(c, z).common(Conc(c, z), suffix=True) == Conc(c, z)
    assert Conc(c, y).common(Conc(c, z), suffix=True) == Conc()
    assert Conc(a, z).common(Conc(b, z), suffix=True) == Conc(z)
    assert Conc(a, zstar).common(Conc(b, z), suffix=True) == Conc()
    assert Conc(a).common(Conc(b), suffix=True) == Conc()


def test_conc_dock() -> None:
    a = Mult(Charclass("A"), ONE)
    b = Mult(Charclass("B"), ONE)
    x = Mult(Charclass("X"), ONE)
    x_twice = Mult(Charclass("X"), Multiplier(Bound(2), Bound(2)))
    yplus = Mult(Charclass("y"), PLUS)
    z = Mult(Charclass("Z"), ONE)

    assert Conc(a, z).dock(Conc(z)) == Conc(a)
    assert Conc(a, b, x, yplus, z).dock(Conc(x, yplus, z)) == Conc(a, b)
    assert Conc(a, b, x, yplus, z).behead(Conc(a, b, x, yplus)) == Conc(z)
    assert Conc(a).dock(Conc()) == Conc(a)

    with pytest.raises(ArithmeticError, match="Can't subtract"):
        Conc(x_twice, yplus, z).behead(Conc(x, yplus))


def test_mult_reduction_easy() -> None:
    assert Conc(Mult(Charclass("a"), ZERO)).reduce() == Conc()



from __future__ import annotations

import pytest

from .bound import INF, Bound


def test_ctor() -> None:
    assert Bound(None) == INF

    Bound(0)
    Bound(1)
    Bound(2)

    with pytest.raises(ValueError):
        Bound(-1)


def test_eq_neq() -> None:
    # pylint: disable=comparison-with-itself
    assert Bound(0) == Bound(0)
    assert INF == INF
    assert Bound(0) != Bound(1)
    assert Bound(0) != INF
    assert Bound(1) == Bound(1)
    assert Bound(None) == INF


def test_eq_neq_heterogeneous() -> None:
    assert Bound(1) != "blah"


def test_comparisons() -> None:
    # pylint: disable=comparison-with-itself
    # pylint: disable=unneeded-not

    assert Bound(0) < Bound(1)
    assert Bound(0) < INF
    assert Bound(1) < INF
    assert not INF < INF


def test_multiplication() -> None:
    assert Bound(0) * Bound(0) == Bound(0)
    assert Bound(0) * Bound(1) == Bound(0)
    assert Bound(0) * Bound(2) == Bound(0)
    assert Bound(0) * Bound(5) == Bound(0)
    assert Bound(0) * INF == Bound(0)

    assert Bound(1) * Bound(5) == Bound(5)
    assert Bound(2) * Bound(5) == Bound(10)
    assert Bound(0) * INF == Bound(0)
    assert Bound(2) * INF == INF
    assert INF * INF == INF
    assert INF * Bound(0) == Bound(0)
    assert Bound(1) * Bound(0) == Bound(0)


def test_addition() -> None:
    assert Bound(0) + Bound(0) == Bound(0)
    assert Bound(0) + Bound(1) == Bound(1)
    assert Bound(0) + Bound(5) == Bound(5)
    assert Bound(0) + INF == INF

    assert Bound(1) + Bound(0) == Bound(1)
    assert Bound(1) + Bound(1) == Bound(2)
    assert Bound(1) + Bound(5) == Bound(6)
    assert Bound(1) + INF == INF

    assert INF + Bound(0) == INF
    assert INF + Bound(1) == INF
    assert INF + INF == INF


def test_subtraction() -> None:
    assert Bound(0) - Bound(0) == Bound(0)
    assert Bound(1) - Bound(0) == Bound(1)
    assert Bound(6) - Bound(4) == Bound(2)
    assert Bound(5) - Bound(5) == Bound(0)

    assert INF - Bound(0) == INF
    assert INF - Bound(1) == INF
    assert INF - Bound(1000) == INF
    assert INF - INF == Bound(0)

    with pytest.raises(ArithmeticError):
        _ = Bound(5) - Bound(6)

    with pytest.raises(ArithmeticError):
        _ = Bound(0) - Bound(1)

    with pytest.raises(ArithmeticError):
        _ = Bound(0) - INF

    with pytest.raises(ArithmeticError):
        _ = Bound(10) - INF


def test_copy() -> None:
    assert INF.copy() == INF

    b = Bound(6)
    assert b.copy() == b


def test_bound_str() -> None:
    assert str(Bound(2)) == "2"

    # pylint: disable-next=compare-to-empty-string
    assert str(INF) == ""


def test_bound() -> None:
    assert min(Bound(0), INF) == Bound(0)
    assert min(Bound(1), INF) == Bound(1)
